"""zyx._graph._ctx"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
import inspect
from typing import Any, Callable, Dict, List, Generic, Self, Type, TypeVar

from pydantic_graph.beta.step import StepContext

from pydantic_ai.toolsets import FunctionToolset
from pydantic_ai.output import NativeOutput
from pydantic_ai.settings import merge_model_settings

from .._processing._messages import (
    parse_context_to_pydantic_ai_messages,
    parse_instructions_as_system_prompt_parts,
)
from .._aliases import (
    PydanticAIAgent,
    PydanticAIAgentResult,
    PydanticAIAgentStream,
    PydanticAIModelSettings,
    PydanticAIModel,
    PydanticAIMessage,
    PydanticAITool,
    PydanticAIBuiltinTool,
    PydanticAIToolset,
    PydanticAIUsage,
    PydanticAIUsageLimits,
    PydanticAISystemPromptPart,
)
from .._utils._outputs import OutputBuilder
from .._utils._strategies._params import SourceStrategy, TargetStrategy
from .._processing._outputs import normalize_output_target
from ..context import Context
from ..attachments import Attachment, AttachmentLike, is_attachment_like
from ..targets import Target

__all__ = (
    "SemanticGraphDeps",
    "SemanticGraphState",
    "SemanticGraphContext",
    "StreamFieldMapping",
    "GraphHooks",
)


_logger = logging.getLogger(__name__)


Deps = TypeVar("Deps")
Output = TypeVar("Output")


class InjectedDeps(Generic[Deps]):
    """Wrapper to expose user deps plus internal graph/runtime deps."""

    def __init__(
        self,
        user: Deps | None,
        internal: Dict[str, Any] | None = None,
    ) -> None:
        self.user = user
        self.internal = internal or {}

    def __getattr__(self, name: str) -> Any:
        if self.user is not None and hasattr(self.user, name):
            return getattr(self.user, name)
        if name in self.internal:
            return self.internal[name]
        raise AttributeError(name)

    def get(self, name: str, default: Any = None) -> Any:
        try:
            return getattr(self, name)
        except AttributeError:
            return default


@dataclass
class GraphHooks:
    on_run_start: Callable[[Any], None] | None = None
    on_run_end: Callable[[Any], None] | None = None
    on_error: Callable[[BaseException, Any], None] | None = None


@dataclass
class SemanticGraphDeps(Generic[Deps, Output]):
    """
    Immutable configuration / dependencies provided to a semantic operation's
    graph for execution.

    The primary purpose of this class is to provide a normalized representation
    of various user-provided parameters in the `pydantic_ai` format, as well as
    contain framework specific parameters.
    """

    agent: PydanticAIAgent
    """The `pydantic_ai.Agent` instance that will be invoked when executing the nodes
    within a semantic operation's graph."""

    target: Type[Output] | Output | Target[Output]
    """The target type or value that is treated as the final output of a semantic operation."""

    source: Any | Type[Any] | None = None
    """A source object that acts as the ground truth or primary context reference for
    specific semantic operations such as `parse` or `query`."""

    confidence: bool = False
    """Whether to include confidence scores in the `Result` of a semantic operation."""

    message_history: List[PydanticAIMessage] = field(default_factory=list)
    """Parsed message history from the `context` and `instructions` parameters of a
    semantic operation, in the `pydantic_ai.ModelMessage` format."""

    instructions: List[PydanticAISystemPromptPart] = field(
        default_factory=list
    )
    """User-provided instructions that are forwarded to the agent, these are parsed into
    `pydantic_ai.SystemPromptPart` objects when being passed to an agent."""

    toolsets: List[PydanticAIToolset] = field(default_factory=list)
    """Unified representation of all user-provided functions, `pydantic_ai` tools, builtin tools
    and toolsets.

    NOTE: All functions, tools and builtin tools are converted into a single `FunctionToolset`
    when passed to an agent."""

    deps: Deps | None = None
    """Reference to `deps` within `pydantic_ai` which can be passed to tools, messages and instructions
    through `pydantic_ai.RunContext`."""

    user_deps: Deps | None = None
    """Original user-provided deps before any internal injection."""

    internal_deps: Dict[str, Any] | None = None
    """Internal framework deps injected into RunContext."""

    usage_limits: PydanticAIUsageLimits | None = None
    """The usage limits to set for a single agent run within the execution of a
    semantic operation's graph."""

    attachments: List[Attachment | AttachmentLike | Any] | None = None
    """A list of attachments provided to the agent.

    An attachment is a piece of content that is provided to the agent in a 'persistent' fashion,
    where it is templated/placed specifically to avoid context rot or loss.
    """

    _context_refs: Context | List[Context] | None = None
    """Reference to any `Context` objects that were found within the `context` parameter
    of a semantic operation."""

    _context_additions: List[PydanticAIMessage] = field(default_factory=list)
    """Non-Context messages provided in the original context parameter."""

    observe: Any | None = None
    """Optional observer for CLI progress and tool events."""

    semantic_renderer: Any | None = None
    """Optional callable to build semantic representations for context updates."""

    hooks: GraphHooks | None = None
    """Optional graph-level lifecycle hooks."""

    source_strategy: SourceStrategy | None = None
    """Unified strategy for handling `source` context rendering."""

    target_strategy: TargetStrategy | None = None
    """Unified strategy for handling target schema and instructions."""

    @classmethod
    def prepare(
        cls,
        model: str | PydanticAIModel | PydanticAIAgent,
        model_settings: PydanticAIModelSettings | None = None,
        context: Any | List[Any] | None = None,
        source: Any | Type[Any] | None = None,
        target: Type[Output] | Output | None = None,
        confidence: bool = False,
        attachments: Attachment
        | AttachmentLike
        | Any
        | List[Attachment | AttachmentLike | Any]
        | None = None,
        instructions: str | Callable | List[str | Callable] | None = None,
        tools: Any | List[Any] | None = None,
        deps: Deps | None = None,
        internal_deps: Dict[str, Any] | None = None,
        inject_internal_deps: bool = False,
        usage_limits: PydanticAIUsageLimits | None = None,
        observe: Any | None = None,
        semantic_renderer: Any | None = None,
        hooks: GraphHooks | None = None,
    ) -> Self:
        """
        Prepares a new instance of `SemanticGraphDeps` through a standardized set of
        parameters shared by semantic operations.
        """
        _context_refs, deps = cls._extract_context_refs(context, deps)

        agent = cls._prepare_agent(
            model=model,
            model_settings=model_settings,
            target=target,
            deps=deps,
        )

        if confidence is True:
            agent = _ensure_confidence_supported_agent(agent)
        instructions_list = cls._normalize_instructions(instructions)
        rendered_instructions = cls._prepare_instructions(
            instructions=instructions_list,
            deps=deps,
            context_refs=_context_refs,
        )
        message_history = parse_context_to_pydantic_ai_messages(context)
        context_additions = cls._extract_context_additions(context)

        toolsets, attachments_list, inject_internal_deps = (
            cls._prepare_tools_and_attachments(
                tools=tools,
                attachments=attachments,
                context_refs=_context_refs,
                inject_internal_deps=inject_internal_deps,
            )
        )

        deps, user_deps, internal = cls._prepare_deps(
            deps=deps,
            internal_deps=internal_deps,
            inject_internal_deps=inject_internal_deps,
            agent=agent,
        )

        observe_handler = cls._prepare_observer(observe)

        return cls(
            agent=agent,
            target=target,
            source=source,
            confidence=confidence,
            message_history=message_history,
            instructions=rendered_instructions,
            toolsets=toolsets,
            deps=deps,
            user_deps=user_deps,
            internal_deps=internal,
            usage_limits=usage_limits,
            attachments=attachments_list,
            _context_refs=_context_refs,
            _context_additions=context_additions,
            observe=observe_handler,
            semantic_renderer=semantic_renderer,
            hooks=hooks,
            source_strategy=SourceStrategy(source)
            if source is not None
            else None,
            target_strategy=TargetStrategy.from_target(target)
            if target is not None
            else None,
        )

    @classmethod
    def _prepare_agent(
        cls,
        *,
        model: str | PydanticAIModel | PydanticAIAgent,
        model_settings: PydanticAIModelSettings | None,
        target: Type[Output] | Output | None,
        deps: Deps | None,
    ) -> PydanticAIAgent:
        agent: PydanticAIAgent | None = None

        if isinstance(model, PydanticAIAgent):
            agent = model
            if model_settings is not None:
                agent.model_settings = merge_model_settings(
                    base=agent.model_settings,
                    overrides=model_settings,
                )
        elif isinstance(model, (PydanticAIModel, str)):
            output_type = None
            if isinstance(target, Target):
                normalized = normalize_output_target(target.target)
                output_type = NativeOutput(
                    outputs=normalized,
                    name=target.name,
                    description=target.description,
                )

            agent = PydanticAIAgent(
                model=model,
                model_settings=model_settings,
                deps_type=type(deps) if deps is not None else None,
                **(
                    {"output_type": output_type}
                    if output_type is not None
                    else {}
                ),
            )  # type: ignore

        if not agent:
            raise ValueError(
                "Invalid 'runner' (model) provided when preparing semantic operation graph dependencies. Accepted formats for a model are:\n"
                "1. A string in the `pydantic_ai` model format (e.g. 'openai:gpt-4o-mini', etc.)\n"
                "2. A `pydantic_ai.models.Model` object\n"
                "3. A `pydantic_ai.Agent` object\n"
            )

        return agent

    @classmethod
    def _extract_context_refs(
        cls, context: Any | List[Any] | None, deps: Deps | None
    ) -> tuple[list[Context], Deps | None]:
        _context_refs: list[Context] = []
        if context:
            ctx_list = context if isinstance(context, list) else [context]
            _context_refs = [
                item for item in ctx_list if isinstance(item, Context)
            ]
            if _context_refs and deps is None:
                last_ctx = _context_refs[-1]
                if getattr(last_ctx, "deps", None) is not None:
                    deps = last_ctx.deps
        return _context_refs, deps

    @classmethod
    def _normalize_instructions(
        cls,
        instructions: str | Callable | List[str | Callable] | None,
    ) -> List[str | Callable]:
        if instructions is None:
            return []
        if not isinstance(instructions, list):
            return [instructions]
        return instructions  # type: ignore[return-value]

    @classmethod
    def _prepare_instructions(
        cls,
        *,
        instructions: List[str | Callable],
        deps: Deps | None,
        context_refs: list[Context],
    ) -> List[PydanticAISystemPromptPart]:
        rendered = parse_instructions_as_system_prompt_parts(
            instructions=instructions,
            deps=deps,
        )
        for ctx in context_refs:
            rendered.extend(ctx.render_instructions())
        return rendered

    @classmethod
    def _extract_context_additions(
        cls, context: Any | List[Any] | None
    ) -> List[PydanticAIMessage]:
        if not context:
            return []
        ctx_list = context if isinstance(context, list) else [context]
        extras = [item for item in ctx_list if not isinstance(item, Context)]
        if not extras:
            return []
        return parse_context_to_pydantic_ai_messages(extras)

    @classmethod
    def _prepare_tools_and_attachments(
        cls,
        *,
        tools: Any | List[Any] | None,
        attachments: Attachment
        | AttachmentLike
        | List[Attachment | AttachmentLike]
        | None,
        context_refs: list[Context],
        inject_internal_deps: bool,
    ) -> tuple[
        List[PydanticAIToolset], List[Attachment | AttachmentLike], bool
    ]:
        ctx_toolsets = []
        for ctx in context_refs:
            ctx_toolsets.extend(ctx.render_toolsets())

        function_tools = []
        toolsets = list(ctx_toolsets)
        attachments_list: List[Attachment | AttachmentLike] = []

        if tools:
            if not isinstance(tools, list):
                tools = [tools]

            for tool in tools:
                if inspect.isfunction(tool):
                    function_tools.append(PydanticAITool(function=tool))
                elif isinstance(tool, (PydanticAIBuiltinTool, PydanticAITool)):
                    function_tools.append(tool)
                elif isinstance(tool, PydanticAIToolset):
                    toolsets.append(tool)
                elif hasattr(tool, "get_toolset"):
                    toolsets.append(tool.get_toolset())
                else:
                    raise ValueError(
                        f"Invalid tool: {tool}. Accepted formats for tools are:\n"
                        "1. A function\n"
                        "2. A `pydantic_ai.tools.Tool` object\n"
                        "3. A `pydantic_ai.toolsets.AbstractToolset` object\n"
                    )

        if function_tools:
            toolsets.append(FunctionToolset(tools=function_tools))

        if attachments:
            if not isinstance(attachments, list):
                attachments = [attachments]
            for attachment in attachments:
                if isinstance(attachment, Attachment) or is_attachment_like(
                    attachment
                ):
                    normalized = attachment
                else:
                    raise ValueError(
                        f"Invalid attachment: {attachment}. Accepted formats for attachments are:\n"
                        "1. An `Attachment`\n"
                        "2. An object with `get_description()`, `get_state_description()`, and `get_toolset()`"
                    )

                attachments_list.append(normalized)
                toolset = normalized.get_toolset()
                if toolset:
                    toolsets.append(toolset)
                    if not inject_internal_deps:
                        inject_internal_deps = True

        return toolsets, attachments_list, inject_internal_deps

    @classmethod
    def _prepare_deps(
        cls,
        *,
        deps: Deps | None,
        internal_deps: Dict[str, Any] | None,
        inject_internal_deps: bool,
        agent: PydanticAIAgent,
    ) -> tuple[Deps | None, Deps | None, Dict[str, Any] | None]:
        user_deps = deps
        internal: Dict[str, Any] | None = internal_deps
        if inject_internal_deps:
            if internal is None:
                internal = {}
            internal.setdefault("agent", agent)
            internal.setdefault("model", agent.model)

        if internal:
            deps = InjectedDeps(user_deps, internal)  # type: ignore[arg-type]

        return deps, user_deps, internal

    @classmethod
    def _prepare_observer(cls, observe: Any | None) -> Any | None:
        observe_handler = None
        if observe:
            from .._utils._observer import Observer

            if observe is True:
                observe_handler = Observer()
            elif isinstance(observe, Observer):
                observe_handler = observe
            else:
                raise ValueError(
                    "Invalid 'observe' value. Expected bool or ObserveHandler."
                )
        return observe_handler


@dataclass
class SemanticGraphState(Generic[Output]):
    """
    Mutable object that is updated / changed on a node-to-node basis as a semantic operation
    is executed.
    """

    output: OutputBuilder[Output]
    """Builder object that tracks the state of the output `target`'s composition
    over the execution of a semantic operation's graph."""

    agent_runs: List[PydanticAIAgentResult] = field(default_factory=list)
    """Accumulation of all `pydantic_ai.AgentRunResult` objects that have been executed by nodes within
    the execution of a semantic operation's graph."""

    usage: PydanticAIUsage = field(default_factory=PydanticAIUsage)
    """Accumulated token usage across all agent runs within the execution of a semantic
    operation."""

    streams: List[PydanticAIAgentStream | None] = field(default_factory=list)
    """Streamed Agent runs that havent been consumed yet by the `Stream` result wrapper."""

    stream_contexts: List[Any] = field(default_factory=list)
    """Context managers that handle the consumption of the content within streams, used for
    cleanup."""

    stream_field_mappings: List["StreamFieldMapping"] = field(
        default_factory=list
    )
    """Metadata about which fields each stream should update.

    Each entry contains:
    - stream_index: Index into self.streams
    - fields: List of field names to update, or None for all fields
    - update_output: Whether to update the output builder
    """

    @classmethod
    def prepare(cls, deps: SemanticGraphDeps[Deps, Output]) -> Self:
        """Prepares a new instance of `SemanticGraphState` using a provided
        `SemanticGraphDeps` instance."""
        if isinstance(deps.target, Target):
            return cls(
                output=OutputBuilder(target=deps.target.target),
            )
        return cls(
            output=OutputBuilder(target=deps.target),
        )


SemanticGraphContext = StepContext[
    SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
]
"""RunContext type passed through the steps of a semantic operation's graph."""


@dataclass
class StreamFieldMapping:
    stream_index: int
    fields: List[str] | None
    update_output: bool


def _ensure_confidence_supported_agent(
    agent: PydanticAIAgent,
) -> PydanticAIAgent:
    """
    Ensures that a given `pydantic_ai.Agent` is configured to return confidence
    scores, when it is active.
    """
    model = agent._get_model(agent.model)
    model_name = model.model_name

    if not model.client or model.client.__class__.__name__ not in (  # type: ignore[attr-defined]
        "OpenAI",
        "AsyncOpenAI",
    ):
        _logger.warning(
            f"Confidence scoring is only supported with OpenAI or OpenAI-like models. "
            f"Model '{model}' may not return log-probabilities, so confidence will be inaccurate."
        )
        return agent
    else:
        if model.__class__.__name__ == "OpenAIChatModel":
            from pydantic_ai.models.openai import (
                OpenAIResponsesModel,
                OpenAIResponsesModelSettings,
            )
            from pydantic_ai.providers.openai import OpenAIProvider

            _logger.debug(
                f"Attempting to switch model class for {model_name} to use OpenAIResponsesModel."
            )
            agent._model = OpenAIResponsesModel(
                model_name=model_name,
                provider=OpenAIProvider(
                    openai_client=model.client  # type: ignore[attr-defined]
                ),
            )

            if not agent.model_settings:
                agent.model_settings = OpenAIResponsesModelSettings(
                    # NOTE: .. openai stop taking peoples data by default....
                    # its quite rude so we also disable it by default
                    openai_logprobs=True,
                    openai_top_logprobs=10,
                    openai_store=False,
                )
            else:
                agent.model_settings = merge_model_settings(
                    base=agent.model_settings,
                    overrides=OpenAIResponsesModelSettings(
                        openai_logprobs=True,
                        openai_top_logprobs=10,
                        openai_store=False
                        if not agent.model_settings.get("openai_store", True)
                        else None,
                    ),
                )
            return agent

    return agent



"""zyx._graph"""

from __future__ import annotations

import asyncio
import threading
import inspect
from types import SimpleNamespace
from typing import Any, Callable, Generic, Sequence, TypeVar

from pydantic_graph.beta.graph import Graph
from pydantic_graph.beta.graph_builder import GraphBuilder
from pydantic_graph.nodes import End

from pydantic_ai import _output as _pydantic_ai_output
from pydantic_ai import _utils as _pydantic_ai_utils
from pydantic_ai.exceptions import ModelRetry

from ._ctx import (
    SemanticGraphDeps,
    SemanticGraphState,
    SemanticGraphContext,
    GraphHooks,
)
from ._nodes import (
    AbstractSemanticNode,
    make_generate_step,
    make_stream_step,
    run_v1_node_chain,
)
from ._requests import SemanticGraphRequestTemplate
from ..targets import Target
from ..result import Result
from ..stream import Stream

__all__ = (
    "SemanticGraph",
    "AbstractSemanticNode",
    "End",
    "make_generate_step",
    "make_stream_step",
    "run_v1_node_chain",
    "SemanticGraphRequestTemplate",
    "SemanticGraphDeps",
    "SemanticGraphState",
    "SemanticGraphContext",
    "GraphHooks",
)


Deps = TypeVar("Deps")
Output = TypeVar("Output")


def _stream_sync_impl(
    self: SemanticGraph[Output], *, exclude_none: bool
) -> Stream[Output]:
    """Run stream() in a dedicated thread so anyio/pydantic_ai context stays isolated."""
    import threading

    result_container: list[
        tuple[Stream[Output], asyncio.AbstractEventLoop]
    ] = []
    error_container: list[BaseException] = []
    ready = threading.Event()

    def _worker() -> None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            stream = loop.run_until_complete(
                self.stream(exclude_none=exclude_none)
            )
            stream._loop = loop
            result_container.append((stream, loop))
        except BaseException as exc:
            error_container.append(exc)
        finally:
            ready.set()
        try:
            loop.run_forever()
        finally:
            loop.close()

    thread = threading.Thread(target=_worker, name="zyx-stream-sync")
    thread.start()
    ready.wait()
    if error_container:
        thread.join()
        raise error_container[0]
    if not result_container:
        thread.join()
        raise RuntimeError("Stream sync worker failed to produce a stream")
    stream, _loop = result_container[0]
    stream._loop_owner_thread = thread
    return stream


def _get_event_loop() -> asyncio.AbstractEventLoop:
    """Get or create event loop for sync operations."""
    try:
        event_loop = asyncio.get_event_loop()
    except RuntimeError:
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
    return event_loop


def _run_coro_sync(coro: Any) -> Any:
    """Run a coroutine from sync code, even if an event loop is already running."""
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(coro)

    result_box: dict[str, Any] = {}
    error_box: dict[str, BaseException] = {}

    def _runner() -> None:
        try:
            result_box["value"] = asyncio.run(coro)
        except BaseException as exc:  # pragma: no cover - re-raised in caller
            error_box["error"] = exc

    thread = threading.Thread(target=_runner, name="zyx-asyncio-run")
    thread.start()
    thread.join()

    if "error" in error_box:
        raise error_box["error"]
    return result_box.get("value")


class SemanticGraph(Generic[Output]):
    """
    Execution wrapper for a semantic operation graph.

    Wraps pydantic_graph's Graph with zyx's state/deps model and provides
    clean execution methods that return properly formatted results.
    """

    def __init__(
        self,
        *,
        steps: Sequence[Callable[..., Any]],
        state: SemanticGraphState[Output],
        deps: SemanticGraphDeps[Deps, Output],
    ) -> None:
        """Create a semantic operation graph.

        Args:
            nodes : The nodes that make up the graph.
            start : The start node of the graph.
            state : The initial state of the graph.
            deps : The dependencies of the graph.
        """
        self._steps = steps
        self._state = state
        self._deps = deps
        self._graph: Graph | None = None

    def prepare_graph(self) -> Graph:
        if self._graph is None:
            builder = GraphBuilder(
                state_type=type(self._state),
                deps_type=type(self._deps),
                input_type=type(None),
                output_type=Any,
            )
            prev = builder.start_node
            for idx, step in enumerate(self._steps):
                step_node = builder.step(call=step, node_id=f"step_{idx}")
                builder.add(builder.edge_from(prev).to(step_node))
                prev = step_node
            builder.add(builder.edge_from(prev).to(builder.end_node))
            self._graph = builder.build()
        return self._graph

    def _auto_update_context(self, res: Result[Output]) -> None:
        """Write messages back to the originating ``Context`` if applicable."""
        ctx_refs = getattr(self._deps, "_context_refs", None) or []
        if not isinstance(ctx_refs, list):
            ctx_refs = [ctx_refs]
        context_additions = getattr(self._deps, "_context_additions", []) or []

        for ctx_ref in ctx_refs:
            if (
                ctx_ref is None
                or not getattr(ctx_ref, "update", True)
                or not res.raw
            ):
                continue

            if context_additions:
                ctx_ref.extend_messages(context_additions)

            semantic_renderer = getattr(self._deps, "semantic_renderer", None)
            if semantic_renderer:
                semantic = semantic_renderer(res, self._state, self._deps)
                if isinstance(semantic, str):
                    res._semantic_message = semantic
                    ctx_ref.add_assistant_message(semantic)
                    continue
                if isinstance(semantic, list):
                    ctx_ref.extend_messages(semantic)
                    continue
                if semantic is not None:
                    ctx_ref.extend_messages([semantic])
                    continue
            ctx_ref.update_from_pydantic_ai_result(res.raw[-1])

    async def run(self) -> Result[Output]:
        """Execute the graph to completion and return the final result."""
        graph = self.prepare_graph()

        cleanup = _install_target_hooks(self._deps)
        _run_graph_start_hooks(self._deps, self._state)

        try:
            if self._state and self._deps:
                await graph.run(
                    state=self._state,
                    deps=self._deps,
                    inputs=None,
                )
        except Exception as e:
            _run_graph_error_hooks(self._deps, e, self._state)
            _run_error_hooks(self._deps, e)
            raise
        finally:
            cleanup()

        res = Result(
            self._state.output.finalize(),
            raw=self._state.agent_runs,
        )
        _run_graph_end_hooks(self._deps, res)
        self._auto_update_context(res)
        return res

    def run_sync(self) -> Result[Output]:
        """Execute the graph to completion and return the final result."""
        graph = self.prepare_graph()

        cleanup = _install_target_hooks(self._deps)
        _run_graph_start_hooks(self._deps, self._state)

        try:
            if self._state and self._deps:
                _run_coro_sync(
                    graph.run(
                        state=self._state,
                        deps=self._deps,
                        inputs=None,
                    )
                )
        except Exception as e:
            _run_graph_error_hooks(self._deps, e, self._state)
            _run_error_hooks(self._deps, e)
            raise
        finally:
            cleanup()

        res = Result(
            self._state.output.finalize(),
            raw=self._state.agent_runs,
        )
        _run_graph_end_hooks(self._deps, res)
        self._auto_update_context(res)
        return res

    async def stream(self, *, exclude_none: bool = False) -> Stream[Output]:
        """Execute the graph with streaming nodes and return a Stream wrapper.

        The graph should contain stream steps that store their streams in
        `ctx.state.streams` during execution. After the graph runs
        to completion, the accumulated streams are collected and wrapped in a
        `Stream` object for consumption.

        Args:
            exclude_none: If ``True``, the resulting ``Stream`` will omit
                ``None``-valued fields when finalizing the output.

        Returns:
            A `Stream[Output]` that can be used to consume the streamed data
            via `stream_text()`, `stream_partial()`, etc.
        """
        graph = self.prepare_graph()

        cleanup = _install_target_hooks(self._deps)
        _run_graph_start_hooks(self._deps, self._state)

        try:
            if self._state and self._deps:
                await graph.run(
                    state=self._state,
                    deps=self._deps,
                    inputs=None,
                )
        except Exception as e:
            _run_graph_error_hooks(self._deps, e, self._state)
            _run_error_hooks(self._deps, e)
            raise
        finally:
            cleanup()

        stream = Stream(
            _builder=self._state.output,
            _streams=self._state.streams,
            _field_mappings=self._state.stream_field_mappings,  # type: ignore[arg-type]
            _stream_contexts=self._state.stream_contexts,
            _exclude_none=exclude_none,
        )
        _run_graph_end_hooks(self._deps, stream)
        return stream

    def stream_sync(self, *, exclude_none: bool = False) -> Stream[Output]:
        """Synchronous wrapper around `stream()`.

        Returns a `Stream[Output]` whose sync iteration methods (`text()`,
        `partial()`, `field()`, `finish()`) can be used from non-async code.

        Runs the graph in a dedicated thread so that anyio cancel scopes and
        contextvars stay within one task/thread, avoiding RuntimeError when
        using pydantic_graph with sync entry points.

        Args:
            exclude_none: Forwarded to ``stream()``.

        Raises:
            RuntimeError: If called from within an already-running async context.
        """
        return _stream_sync_impl(self, exclude_none=exclude_none)


def _install_target_hooks(
    deps: SemanticGraphDeps[Any, Any],
) -> Callable[[], None]:
    target = deps.target
    if not isinstance(target, Target):
        return lambda: None

    validators = _build_target_output_validators(target)
    if not validators:
        return lambda: None

    agent = deps.agent
    start = len(agent._output_validators)
    agent._output_validators.extend(validators)

    def _cleanup() -> None:
        del agent._output_validators[start:]

    return _cleanup


def _build_target_output_validators(
    target: Target[Any],
) -> list[_pydantic_ai_output.OutputValidator[Any, Any]]:
    field_hooks = target._field_hooks
    complete_hooks = target._prebuilt_hooks.get("complete", [])

    if not field_hooks and not complete_hooks:
        return []

    async def _validator(ctx, data):
        value = data

        # Field hooks
        if field_hooks:
            if hasattr(value, "model_copy"):
                updates = {}
                for field, hooks in field_hooks.items():
                    if field == "__self__":
                        value, did_update = await _apply_hooks(
                            hooks, ctx, value, field
                        )
                        if not did_update:
                            value = data
                        continue
                    if not hasattr(value, field):
                        continue
                    current = getattr(value, field)
                    if current is None:
                        continue
                    updated, did_update = await _apply_hooks(
                        hooks, ctx, current, field
                    )
                    if did_update:
                        updates[field] = updated
                if updates:
                    value = value.model_copy(update=updates)
            elif isinstance(value, dict):
                for field, hooks in field_hooks.items():
                    if field == "__self__":
                        value, did_update = await _apply_hooks(
                            hooks, ctx, value, field
                        )
                        if not did_update:
                            value = data
                        continue
                    if field not in value or value[field] is None:
                        continue
                    updated, did_update = await _apply_hooks(
                        hooks, ctx, value[field], field
                    )
                    if did_update:
                        value[field] = updated
            else:
                if "__self__" in field_hooks:
                    value, did_update = await _apply_hooks(
                        field_hooks["__self__"], ctx, value, "__self__"
                    )
                    if not did_update:
                        value = data

        # Complete hooks
        if complete_hooks:
            value, did_update = await _apply_hooks(
                complete_hooks, ctx, value, "__complete__"
            )
            if not did_update:
                value = data

        return value

    return [_pydantic_ai_output.OutputValidator(_validator)]


async def _apply_hooks(hooks, ctx, value, field_name: str):
    current = value
    did_update = False
    for fn, retry, update in hooks:
        try:
            takes_ctx = len(inspect.signature(fn).parameters) > 1
            if takes_ctx:
                args = (ctx, current)
            else:
                args = (current,)

            if _pydantic_ai_utils.is_async_callable(fn):
                next_value = await fn(*args)
            else:
                next_value = fn(*args)
            current = next_value
            if update:
                did_update = True
        except Exception as e:
            if retry:
                raise ModelRetry(
                    f"Target hook failed for '{field_name}': {e}"
                ) from e
            raise
    return current, did_update


def _run_error_hooks(
    deps: SemanticGraphDeps[Any, Any], error: BaseException
) -> None:
    target = deps.target
    if not isinstance(target, Target):
        return

    hooks = target._prebuilt_hooks.get("error", [])
    if not hooks:
        return

    for fn, _retry, _update in hooks:
        try:
            takes_ctx = len(inspect.signature(fn).parameters) > 1
            if takes_ctx:
                fn(deps)
            else:
                fn(deps)
        except Exception:
            pass


def _run_graph_start_hooks(
    deps: SemanticGraphDeps[Any, Any], state: SemanticGraphState[Any]
) -> None:
    hooks: GraphHooks | None = getattr(deps, "hooks", None)
    if hooks is None or hooks.on_run_start is None:
        return
    try:
        hooks.on_run_start(SimpleNamespace(state=state, deps=deps))
    except Exception:
        pass


def _run_graph_end_hooks(
    deps: SemanticGraphDeps[Any, Any], result: Any
) -> None:
    hooks: GraphHooks | None = getattr(deps, "hooks", None)
    if hooks is None or hooks.on_run_end is None:
        return
    try:
        hooks.on_run_end(result)
    except Exception:
        pass


def _run_graph_error_hooks(
    deps: SemanticGraphDeps[Any, Any],
    error: BaseException,
    state: SemanticGraphState[Any],
) -> None:
    hooks: GraphHooks | None = getattr(deps, "hooks", None)
    if hooks is None or hooks.on_error is None:
        return
    try:
        hooks.on_error(
            error,
            SimpleNamespace(state=state, deps=deps),
        )
    except Exception:
        pass



"""zyx._aliases

Provides interface-level aliases that re-export various components and types from the
`pydantic_ai` library for more coherent usage throughout `zyx`.
"""

from __future__ import annotations

from typing import TypeAlias

from pydantic_ai import (
    _function_schema as _pydantic_ai_function_schema,
    agent as _pydantic_ai_agent,
    messages as _pydantic_ai_messages,
    models as _pydantic_ai_models,
    result as _pydantic_ai_result,
    run as _pydantic_ai_run,
    tools as _pydantic_ai_tools,
    toolsets as _pydantic_ai_toolsets,
    settings as _pydantic_ai_settings,
    usage as _pydantic_ai_usage,
    RunContext as PydanticAIRunContext,
)

__all__ = (
    "PydanticAIRunContext",
    "PydanticAIAgent",
    "PydanticAIAgentResult",
    "PydanticAIAgentStream",
    "PydanticAIInstructions",
    "PydanticAIMessage",
    "PydanticAIModel",
    "PydanticAIModelName",
    "PydanticAIModelSettings",
    "PydanticAIModelRequest",
    "PydanticAIModelResponse",
    "PydanticAISystemPromptPart",
    "PydanticAIUserPromptPart",
    "PydanticAIUserContent",
    "PydanticAITool",
    "PydanticAIBuiltinTool",
    "PydanticAIToolset",
    "PydanticAIUsage",
    "PydanticAIUsageLimits",
    "_pydantic_ai_function_schema",
)


PydanticAIAgent: TypeAlias = _pydantic_ai_agent.Agent


PydanticAIAgentResult: TypeAlias = _pydantic_ai_run.AgentRunResult


PydanticAIAgentStream: TypeAlias = _pydantic_ai_result.StreamedRunResult


PydanticAIInstructions: TypeAlias = _pydantic_ai_agent.Instructions


PydanticAIMessage: TypeAlias = _pydantic_ai_messages.ModelMessage


PydanticAIModel: TypeAlias = _pydantic_ai_models.Model


PydanticAIModelName: TypeAlias = _pydantic_ai_models.KnownModelName


PydanticAIModelSettings: TypeAlias = _pydantic_ai_settings.ModelSettings


PydanticAIModelRequest: TypeAlias = _pydantic_ai_messages.ModelRequest


PydanticAIModelResponse: TypeAlias = _pydantic_ai_messages.ModelResponse


PydanticAISystemPromptPart: TypeAlias = _pydantic_ai_messages.SystemPromptPart


PydanticAIUserPromptPart: TypeAlias = _pydantic_ai_messages.UserPromptPart


PydanticAIUserContent: TypeAlias = _pydantic_ai_messages.UserContent


PydanticAITool: TypeAlias = _pydantic_ai_tools.Tool


PydanticAIBuiltinTool: TypeAlias = _pydantic_ai_tools.AbstractBuiltinTool


PydanticAIToolset: TypeAlias = _pydantic_ai_toolsets.AbstractToolset


PydanticAIUsage: TypeAlias = _pydantic_ai_usage.RunUsage


PydanticAIUsageLimits: TypeAlias = _pydantic_ai_usage.UsageLimits



"""zyx._strategies._attachments"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal

from pydantic_ai.toolsets import FunctionToolset
from pydantic_ai import RunContext

from ._abstract import AbstractStrategy
from ..._processing._multimodal import (
    MultimodalContentOrigin,
    MultimodalContentMediaType,
    render_multimodal_source_as_text,
)
from ..._processing._toon import object_as_text


@dataclass
class AnchorEdit:
    start_anchor: str
    end_anchor: str
    replacement: str


def _apply_anchor_edit(
    text: str, *, start_anchor: str, end_anchor: str, replacement: str
) -> str:
    start_idx = text.find(start_anchor)
    if start_idx == -1:
        return text
    end_idx = text.find(end_anchor, start_idx + len(start_anchor))
    if end_idx == -1:
        return text
    end_idx += len(end_anchor)
    return text[:start_idx] + replacement + text[end_idx:]


@dataclass
class FileStrategy(AbstractStrategy):
    path: Path
    writeable: bool
    confirm: bool
    max_chars: int

    @property
    def kind(self) -> Literal["file"]:
        return "file"

    def get_description(self) -> str:
        return f"File at {self.path}"

    def get_state_description(self) -> str:
        origin = MultimodalContentOrigin.classify(self.path)
        media_type = MultimodalContentMediaType.classify(self.path, origin)
        text = render_multimodal_source_as_text(self.path, origin, media_type)
        if len(text) > self.max_chars:
            return text[: self.max_chars] + "\n\n[TRUNCATED]"
        return text

    def get_toolset(self) -> FunctionToolset | None:
        if not self.writeable:
            return None
        toolset = FunctionToolset()

        @toolset.tool
        def read_file() -> str:
            """Read the current contents of the file."""
            return self.get_state_description()

        @toolset.tool
        def write_file(
            edits: list[AnchorEdit] | None = None,
            content: str | None = None,
        ) -> str:
            """Apply anchor-based edits to the file, or replace it if content is provided."""
            if self.confirm:
                raise RuntimeError("File writes require confirmation.")
            if edits:
                text = self.path.read_text(encoding="utf-8")
                updated = text
                for edit in edits:
                    updated = _apply_anchor_edit(
                        updated,
                        start_anchor=edit.start_anchor,
                        end_anchor=edit.end_anchor,
                        replacement=edit.replacement,
                    )
                self.path.write_text(updated, encoding="utf-8")
                return "ok"
            if content is not None:
                self.path.write_text(content, encoding="utf-8")
            return "ok"

        @toolset.tool
        def append_file(content: str) -> str:
            """Append content to the file."""
            if self.confirm:
                raise RuntimeError("File writes require confirmation.")
            with self.path.open("a", encoding="utf-8") as f:
                f.write(content)
            return "ok"

        return toolset


@dataclass
class ObjectValueStrategy(AbstractStrategy):
    source: Any

    @property
    def kind(self) -> Literal["object"]:
        return "object"

    def get_description(self) -> str:
        return f"Object of type {type(self.source).__name__}"

    def get_state_description(self) -> str:
        return object_as_text(self.source)

    def get_toolset(self) -> FunctionToolset:
        toolset = FunctionToolset()

        @toolset.tool
        def read_object() -> str:
            """Read the current state of the attached object."""
            return self.get_state_description()

        @toolset.tool
        async def edit_object(ctx: RunContext[Any], instructions: str) -> str:
            """Edit the attached object using natural language instructions."""
            from ...operations.edit import aedit

            agent = None
            if hasattr(ctx.deps, "internal"):
                agent = ctx.deps.internal.get("agent")
                if agent is None:
                    agent = ctx.deps.internal.get("model")
            if agent is None:
                agent = getattr(ctx, "model", None)
            if agent is None:
                raise RuntimeError(
                    "No agent or model available in deps for edit_object."
                )

            model = agent
            if hasattr(agent, "model"):
                model = agent.model

            result = await aedit(
                target=self.source,
                context=instructions,
                model=model,
                selective=False,
                deps=getattr(ctx.deps, "user", ctx.deps),
            )
            if result.output is self.source:
                return self.get_state_description()
            if isinstance(self.source, dict) and isinstance(
                result.output, dict
            ):
                self.source.clear()
                self.source.update(result.output)
                return self.get_state_description()
            self.source = result.output
            return self.get_state_description()

        return toolset



"""zyx.tools.memory.qdrant"""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Dict
import warnings

try:
    from qdrant_client import AsyncQdrantClient  # type: ignore[unresolved-import]
    from qdrant_client.http import models  # type: ignore[unresolved-import]
except ImportError:
    raise ImportError(
        "To use the `qdrant` memory provider, you must first install the `qdrant-client` library.\n"
        "You can install it by using one of the following commands:\n"
        "```bash\n"
        "pip install zyx[qdrant]\n"
        "pip install 'qdrant-client[fastembed]'\n"
        "```"
    )

from . import MemoryProvider

warnings.filterwarnings(
    "ignore",
    message=r".*`add` method has been deprecated.*",
    category=UserWarning,
    module=r".*qdrant_client\.common\.client_warnings.*",
)
warnings.filterwarnings(
    "ignore",
    message=r".*`query` method has been deprecated.*",
    category=UserWarning,
    module=r".*qdrant_client\.common\.client_warnings.*",
)

__all__ = ("QdrantMemoryProvider",)


@dataclass(init=False)
class QdrantMemoryProvider(MemoryProvider):
    """
    Memory provider that utilizes `qdrant-client` and `fastembed` for
    embeddings. Uses a local persisted store by default.
    """

    @property
    def client(self) -> AsyncQdrantClient | None:
        """The `qdrant_client.AsyncQdrantClient` instance for this memory provider."""
        return getattr(self, "_client", None)

    @property
    def path(self) -> str | Path | None:
        """The path to the Qdrant database for this memory provider."""
        if self.client is None:
            return None
        return getattr(self.client, "path", None) or getattr(
            self.client, "location", None
        )

    @property
    def collection_name(self) -> str:
        """The name of the collection to use. May contain a `{key}` placeholder."""
        return getattr(self, "_collection_name", "memory-{key}")

    @collection_name.setter
    def collection_name(self, value: str) -> None:
        self._collection_name = value

    def __init__(
        self,
        client: AsyncQdrantClient | Path | str | None = None,
        path: str | Path | None = None,
        collection_name: str = "memory-{key}",
    ):
        """
        Initialize a new `QdrantMemoryProvider` instance.

        Args:
            client : `AsyncQdrantClient` | `Path` | `str` | None
                An existing `AsyncQdrantClient`, or a path for local persistence.
                If None, uses `path` or default `Path.cwd() / "memory/qdrant"`.
            path : `str` | `Path` | None
                Directory path for local Qdrant storage. Used only when `client`
                is None. Defaults to current working directory + `memory/qdrant`.
            collection_name : `str`
                Collection name template; `{key}` is replaced by the memory key.
        """
        if client is not None and not isinstance(client, AsyncQdrantClient):
            path = Path(client) if isinstance(client, str) else client
        if path is None:
            path = Path.cwd() / "memory/qdrant"
        if not isinstance(client, AsyncQdrantClient):
            self._client = AsyncQdrantClient(path=str(path))
        else:
            self._client = client
        self._collection_name = collection_name

    async def get_collection(self, key: str) -> models.CollectionInfo:
        """Return collection info for the given memory key."""
        name = self.collection_name.format(key=key)

        if self.client is not None and hasattr(self.client, "get_collection"):
            return await self.client.get_collection(collection_name=name)

    async def add(self, key: str, content: str) -> str:
        memory_id = str(uuid.uuid4())
        collection_name = self.collection_name.format(key=key)

        if self.client is not None and hasattr(self.client, "add"):
            ids = await self.client.add(
                collection_name=collection_name,
                documents=[content],
                metadata=[{"id": memory_id}],
                ids=[memory_id],
            )
        return str(ids[0]) if ids else memory_id

    async def delete(self, key: str, id: str) -> None:
        collection_name = self.collection_name.format(key=key)

        if self.client is not None and hasattr(self.client, "delete"):
            await self.client.delete(
                collection_name=collection_name,
                points_selector=[id],
            )

    async def search(
        self, key: str, query: str, n: int = 20
    ) -> Dict[str, str]:
        collection_name = self.collection_name.format(key=key)

        if self.client is not None and hasattr(self.client, "query"):
            results = await self.client.query(
                collection_name=collection_name,
                query_text=query,
                limit=n,
            )
        return {str(r.id): r.document for r in results}



"""zyx.attachments"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Protocol, TypeGuard, runtime_checkable
from pydantic_ai.toolsets import FunctionToolset
from pydantic_ai.messages import UserPromptPart

from ._aliases import PydanticAIModelRequest, PydanticAIToolset
from ._processing._multimodal import (
    MultimodalContentOrigin,
    MultimodalContentMediaType,
    render_multimodal_source_as_text,
    render_multimodal_source_as_user_content,
    render_multimodal_source_as_description,
)
from ._utils._strategies._attachments import (
    FileStrategy,
    ObjectValueStrategy,
)

__all__ = (
    "Attachment",
    "AttachmentLike",
    "attach",
    "is_attachment_like",
    "paste",
)


class Attachment:
    """
    An `Attachment` is defined best as 2 sources of content:

    1. A piece of content that can be `pasted` to represent textual
    content, documents or multimodal (image, audio, video) content within
    the `context`, or attachments provided to agents and llms during an operation.

    2. A piece of content that can be `attached` to represent a
    pythonic object or external source that can be queried/edited/etc. by
    an agent or model.

    This supports the following source types:

    - A `Path` object representing a local file
    - A string representing a URL to a text or multimodal file/html page
    - A bytes object representing raw binary data
    - An object representing a pydantic model or other object that can be
    encoded to text using the TOON format
    - And much more...
    """

    def __init__(
        self,
        source: Any,
        *,
        name: str | None = None,
        interactive: bool = False,
        writeable: bool = False,
        confirm: bool = True,
        max_chars: int = 8000,
    ) -> None:
        self.source = source
        self.name = name
        self.interactive = interactive
        self._origin: MultimodalContentOrigin | None = None
        self._media_type: MultimodalContentMediaType | None = None

        self._origin = MultimodalContentOrigin.classify(self.source)
        self._media_type = MultimodalContentMediaType.classify(
            self.source, self._origin
        )
        path = Path(source) if isinstance(source, (str, Path)) else None
        if path is not None and path.exists() and path.is_file():
            self._strategy = FileStrategy(
                path=path,
                writeable=interactive or writeable,
                confirm=confirm,
                max_chars=max_chars,
            )
        else:
            self._strategy = ObjectValueStrategy(source)

    def get_description(self) -> str:
        return self._strategy.get_description()

    def get_state_description(self) -> str | None:
        return self._strategy.get_state_description()

    def get_toolset(self) -> FunctionToolset | None:
        if not self.interactive:
            return None
        return self._strategy.get_toolset()

    @property
    def message(self) -> PydanticAIModelRequest | None:
        if self._origin is None or self._media_type is None:
            return None
        content = render_multimodal_source_as_user_content(
            self.source,
            self._origin,
            self._media_type,
        )
        return PydanticAIModelRequest(
            parts=[UserPromptPart(content=[content])]
        )

    @property
    def description(self) -> str:
        if self._origin is None or self._media_type is None:
            return self.get_description()
        return render_multimodal_source_as_description(
            self.source,
            self._origin,
            self._media_type,
        )

    @property
    def text(self) -> str:
        if self._origin is None or self._media_type is None:
            return self.get_state_description() or ""
        return render_multimodal_source_as_text(
            self.source,
            self._origin,
            self._media_type,
        )


@runtime_checkable
class AttachmentLike(Protocol):
    def get_description(self) -> str: ...

    def get_state_description(self) -> str | None: ...

    def get_toolset(self) -> PydanticAIToolset | None: ...

    @property
    def message(self) -> PydanticAIModelRequest | None: ...

    @property
    def name(self) -> str | None: ...


def is_attachment_like(value: Any) -> TypeGuard[AttachmentLike]:
    return (
        hasattr(value, "get_description")
        and hasattr(value, "get_state_description")
        and hasattr(value, "get_toolset")
    )


def attach(
    source: Any,
    *,
    name: str | None = None,
    writeable: bool = True,
    confirm: bool = True,
    max_chars: int = 8000,
) -> Attachment:
    """
    Attach an attachment from a source object or reference.

    This creates an interactive interface in which a model can use to query, mutate and otherwise
    interact with the attachment.

    Args:
        source (Any): The source object or reference to attach.
        name (str | None): The name of the attachment. Defaults to None.
        writeable (bool): Whether the attachment is writeable. Defaults to True.
        confirm (bool): Whether to confirm the attachment. Defaults to True.
        max_chars (int): The maximum number of characters to attach. Defaults to 8000.

    Returns:
        Attachment: The attachment object.
    """
    return Attachment(
        source,
        name=name,
        interactive=True,
        writeable=writeable,
        confirm=confirm,
        max_chars=max_chars,
    )


def paste(
    source: Any,
    *,
    name: str | None = None,
    max_chars: int = 8000,
) -> Attachment:
    """
    'Paste' an attachment from a source object or reference.

    A source can include a variety of data types, including:
    - A string representing a URL to a web page, multimodal content such as PDFs, images, audio, video, etc.
    - A local file path to a file on the filesystem
    - A bytes object representing raw binary data
    - A Pythonic object that can be encoded to text using the TOON format
    - And more...

    Args:
        source (Any): The source object or reference to paste.
        name (str | None): The name of the attachment. Defaults to None.
        max_chars (int): The maximum number of characters to paste. Defaults to 8000.

    Examples:
    ```python
    from zyx import paste
    attachment = paste("https://zyx.hammad.app")
    print(attachment.text)
    ```

    Returns:
        Attachment: The attachment object.
    """
    return Attachment(
        source,
        name=name,
        interactive=False,
        max_chars=max_chars,
    )



"""zyx.resources.code

Provides a Code resource that uses Monty to safely execute Python code written
by AI agents.
"""

from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List

from pydantic_ai.toolsets import FunctionToolset
import pydantic_monty


__all__ = ("Code",)


@dataclass(init=False)
class Code:
    """
    A resource that allows safely executing Python code using Monty, a minimal,
    secure Python interpreter written in Rust.

    This resource enables AI agents to write and execute Python code without
    the overhead of container-based sandboxes, while maintaining strict security
    controls over filesystem, network, and environment access.
    """

    script_name: str = field(default="code.py")
    """The name of the script file for error reporting and type checking."""

    type_check: bool = field(default=True)
    """Whether to enable type checking for the code."""

    type_check_stubs: str | None = field(default=None)
    """Optional type definitions/stubs for type checking."""

    external_functions: Dict[str, Callable] = field(default_factory=dict)
    """Dictionary mapping function names to their implementations that can be
    called from the executed code."""

    def __init__(
        self,
        *,
        confirm: bool = False,
        script_name: str = "code.py",
        type_check: bool = True,
        type_check_stubs: str | None = None,
        external_functions: Dict[str, Callable] | None = None,
    ) -> None:
        self.confirm = confirm
        self.script_name = script_name
        self.type_check = type_check
        self.type_check_stubs = type_check_stubs
        self.external_functions = external_functions or {}

    def get_description(self) -> str:
        return (
            f"Code execution environment using Monty. "
            f"Supports executing Python code safely with controlled access to "
            f"external functions: {', '.join(self.external_functions.keys()) or 'none'}."
        )

    def get_state_description(self) -> str:
        return (
            f"Code resource ready for execution. "
            f"External functions available: {', '.join(self.external_functions.keys()) or 'none'}."
        )

    def _create_monty(
        self,
        code: str,
        inputs: List[str] | None = None,
    ) -> pydantic_monty.Monty:
        """Create a Monty instance for the given code."""
        return pydantic_monty.Monty(
            code,
            inputs=inputs or [],
            external_functions=list(self.external_functions.keys()),
            script_name=self.script_name,
            type_check=self.type_check,
            type_check_stubs=self.type_check_stubs,
        )

    async def _run_async(
        self,
        code: str,
        inputs: Dict[str, Any] | None = None,
    ) -> Any:
        """Execute code asynchronously and return the result."""
        m = self._create_monty(
            code, inputs=list(inputs.keys()) if inputs else None
        )
        return await pydantic_monty.run_monty_async(
            m,
            inputs=inputs if inputs else None,
            external_functions=self.external_functions,
        )

    def _run_sync(
        self,
        code: str,
        inputs: Dict[str, Any] | None = None,
    ) -> Any:
        """Execute code synchronously and return the result."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            raise RuntimeError(
                "Cannot call synchronous 'run_code' inside an async context. "
                "Use 'await async_run_code' instead."
            )
        else:
            return asyncio.run(self._run_async(code, inputs))

    async def async_run_code(
        self,
        code: str,
        inputs: Dict[str, Any] | None = None,
    ) -> Any:
        """
        Execute Python code asynchronously using Monty and return the result.

        Args:
            code: The Python code to execute.
            inputs: Optional dictionary of input variables to pass to the code.

        Returns:
            The result of executing the code.
        """
        return await self._run_async(code, inputs)

    def run_code(
        self,
        code: str,
        inputs: Dict[str, Any] | None = None,
    ) -> Any:
        """
        Execute Python code synchronously using Monty and return the result.

        Args:
            code: The Python code to execute.
            inputs: Optional dictionary of input variables to pass to the code.

        Returns:
            The result of executing the code.
        """
        return self._run_sync(code, inputs)

    def get_toolset(self) -> FunctionToolset:
        """Get the toolset for interacting with this code resource."""
        toolset = FunctionToolset()

        @toolset.tool
        def execute_code(
            code: str,
            inputs: Dict[str, Any] | None = None,
        ) -> Any:
            """
            Execute Python code safely using Monty.

            The code can call external functions that have been registered with
            this Code resource. The code runs in a secure sandbox with no access
            to filesystem, network, or environment variables unless explicitly
            provided via external functions.

            Args:
                code: The Python code to execute.
                inputs: Optional dictionary of input variables to pass to the code.

            Returns:
                The result of executing the code.
            """
            return self.run_code(code, inputs)

        return toolset



"""zyx.tools.memory

Provides Vector Database based tools that can be used by an agent or
model to store and query memories.

NOTE:
The implementation structure of this resource is essentially a rewrite of the
`Memory` system within the `marvin` library. You can find their implementation
here:

[Marvin Memory](https://github.com/PrefectHQ/marvin/blob/main/src/marvin/memory/memory.py)
"""

from __future__ import annotations

import asyncio
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Literal, TypeAlias

from pydantic_ai.toolsets import FunctionToolset

from ..._aliases import PydanticAITool, PydanticAIToolset

__all__ = (
    "MemoryProvider",
    "Memory",
)


def sanitize_memory_key(key: str) -> str:
    # Remove any characters that are not alphanumeric or underscore
    return re.sub(r"[^a-zA-Z0-9_]", "", key)


MemoryProviderName: TypeAlias = Literal[
    "chroma/persistent",
    "chroma/ephemeral",
    "qdrant",
]


@dataclass(kw_only=True)
class MemoryProvider(ABC):
    """
    Abstract base class for a memory provider.
    """

    def configure(self, key: str) -> None:
        """
        Configure this memory provider with a given identifier or key.
        """

    @abstractmethod
    async def add(self, key: str, content: str) -> str:
        """
        Add some content to the index of memories, which will return it's
        memory id.
        """

    @abstractmethod
    async def delete(self, key: str, id: str) -> None:
        """Delete a memory by it's ID from the index of memories."""

    @abstractmethod
    async def search(
        self, key: str, query: str, n: int = 20
    ) -> Dict[str, str]:
        """Search the index for `n` memories using a string query."""


@dataclass(init=False)
class Memory:
    """
    A tool that can be used by a model/agent to store, query and retrieve
    memories that are stored in a vector database.

    A `Memory` resource can be configured based on a given `MemoryProvider`.
    """

    def __init__(
        self,
        key: str,
        provider: MemoryProvider | MemoryProviderName = "chroma/persistent",
        instructions: str | None = None,
        auto: bool = False,
        writeable: bool = True,
    ):
        """
        Create a new `Memory` resource with a given key, provider, instructions and auto mode.

        Args:
            key : `str`
                The key to use for this memory resource.
            provider : `MemoryProvider` | `MemoryProviderName`
                The provider to use for this memory resource.
            instructions : `str` | None
                The instructions to use for this memory resource.
            auto : `bool`
                Whether to automatically add memories when a semantic operation is executed, using
                the context of the operation as the content to use for the query.
            writeable : `bool`
                Whether the memory can be written to.
        """
        key = sanitize_memory_key(key)

        self.writeable = writeable
        self.key = key
        self._provider = get_memory_provider(provider)
        self.instructions = instructions
        self.auto = auto

    def __hash__(self) -> int:
        return id(self)

    @property
    def provider(self) -> MemoryProvider:
        """The `MemoryProvider` instance that will be used to store and query memories."""
        return self._provider  # type: ignore[return]

    async def async_add(self, content: str) -> str:
        """
        Add some content to the index of memories, which will return it's
        memory id.
        """
        return await self.provider.add(self.key, content)

    def add(self, content: str) -> str:
        """Add some content to the index of memories, which will return its
        memory id."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            raise RuntimeError(
                "Cannot call synchronous 'add' inside an async context. Use 'await async_add'."
            )
        else:
            return asyncio.run(self.async_add(content))

    async def async_delete(self, id: str) -> None:
        """Delete a memory by it's ID from the index of memories."""
        return await self.provider.delete(self.key, id)

    def delete(self, id: str) -> None:
        """Delete a memory by it's ID from the index of memories."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            raise RuntimeError(
                "Cannot call synchronous 'delete' inside an async context. Use 'await async_delete'."
            )
        else:
            return asyncio.run(self.async_delete(id))

    async def async_search(self, query: str, n: int = 20) -> Dict[str, str]:
        """Search the index for `n` memories using a string query."""
        return await self.provider.search(self.key, query, n)

    def search(self, query: str, n: int = 20) -> Dict[str, str]:
        """Search the index for `n` memories using a string query."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None
        if loop and loop.is_running():
            raise RuntimeError(
                "Cannot call synchronous 'search' inside an async context. Use 'await async_search'."
            )
        else:
            return asyncio.run(self.async_search(query, n))

    def get_toolset(self) -> PydanticAIToolset:
        write_tools = [
            PydanticAITool(
                function=self.add,
                name=f"add_memory_{self.key}",
                description="Add a memory to the index of memories.",
            ),
            PydanticAITool(
                function=self.delete,
                name=f"delete_memory_{self.key}",
                description="Delete a memory by it's ID from the index of memories.",
            ),
        ]
        read_tools = [
            PydanticAITool(
                function=self.search,
                name=f"search_memory_{self.key}",
                description="Search the index of memories for a given query.",
            ),
        ]

        if self.writeable:
            return FunctionToolset(tools=[*write_tools, *read_tools])
        else:
            return FunctionToolset(tools=read_tools)

    def get_description(self) -> str:
        if not self.instructions:
            return ""
        return f"[MEMORY : {self.key}]\n\n{self.instructions}"

    def get_state_description(self) -> str:
        return ""


def get_memory_provider(
    provider: MemoryProvider | MemoryProviderName,
) -> MemoryProvider | None:
    """
    Get a `MemoryProvider` instance based on a given name or pre-configured instance.
    """
    if isinstance(provider, MemoryProvider):
        return provider

    elif provider.startswith("chroma/"):
        from .chroma import ChromaMemoryProvider

        if provider == "chroma/persistent":
            return ChromaMemoryProvider(client="persistent")
        elif provider == "chroma/ephemeral":
            return ChromaMemoryProvider(client="ephemeral")

    elif provider == "qdrant":
        from .qdrant import QdrantMemoryProvider

        return QdrantMemoryProvider()

    else:
        raise ValueError(f"Invalid memory provider name: {provider}")


def memories(
    key: str = "memories",
    provider: MemoryProvider | MemoryProviderName = "qdrant",
    instructions: str | None = None,
    auto: bool = False,
    writeable: bool = True,
) -> Memory:
    """
    Create or retrieve a `Memory` resource with based on a given key (collection
    name) and provider.

    Args:
        key : `str`
            The key to use for this memory resource.
        provider : `MemoryProvider` | `MemoryProviderName`
            The provider to use for this memory resource.
        instructions : `str` | None
            The instructions to use for this memory resource.
        auto : `bool`
            Whether to automatically add memories when a semantic operation is executed, using
            the context of the operation as the content to use for the query.
        writeable : `bool`
            Whether the memory can be written to.

    Returns:
        `Memory`
            A `Memory` resource with the given key, provider, instructions and auto mode.
    """
    return Memory(
        key=key,
        provider=provider,
        instructions=instructions,
        auto=auto,
        writeable=writeable,
    )



"""zyx.tools.tool"""

import asyncio
from typing import (
    Any,
    Callable,
    Generic,
    ParamSpec,
    TypeVar,
    cast,
)

from pydantic_ai._run_context import get_current_run_context

from .._aliases import (
    _pydantic_ai_function_schema,
    _pydantic_ai_tools,
    PydanticAIRunContext,
)

__all__ = ("Tool",)


Deps = TypeVar("Deps")


ToolReturn = TypeVar("ToolReturn")


ToolParams = ParamSpec("ToolParams")


class Tool(
    _pydantic_ai_tools.Tool[Deps], Generic[Deps, ToolParams, ToolReturn]
):
    """A tool that can be given to a model or an agent for tool calling and execution during
    a semantic operation.

    This is a wrapper around the `pydantic_ai.tools.Tool` class that allows for the use of
    the __call__ method to invoke the tool function directly.
    """

    def __init__(
        self,
        function: Callable[ToolParams, ToolReturn]
        | Callable[[PydanticAIRunContext[Deps], ToolParams], ToolReturn],
        *,
        takes_ctx: bool | None = None,
        max_retries: int | None = None,
        name: str | None = None,
        description: str | None = None,
        prepare: _pydantic_ai_tools.ToolPrepareFunc[
            _pydantic_ai_tools.ToolAgentDepsT
        ]
        | None = None,
        docstring_format: _pydantic_ai_tools.DocstringFormat = "auto",
        require_parameter_descriptions: bool = False,
        schema_generator: type[
            _pydantic_ai_tools.GenerateJsonSchema
        ] = _pydantic_ai_tools.GenerateToolJsonSchema,
        strict: bool | None = None,
        sequential: bool = False,
        requires_approval: bool = False,
        metadata: dict[str, Any] | None = None,
        timeout: float | None = None,
        function_schema: _pydantic_ai_function_schema.FunctionSchema
        | None = None,
    ) -> None:
        """
        Initialize a new `Tool` instance.

        Args:
            function: Callable[ToolParams, ToolReturn] | Callable[[PydanticAIRunContext[Deps], ToolParams], ToolReturn]
                The function to wrap as a `pydantic_ai.tools.Tool`.
            takes_ctx: bool | None = None
                Whether the tool takes a `PydanticAIRunContext` as its first argument.
            max_retries: int | None = None
                The maximum number of retries for the tool.
            name: str | None = None
                The name of the tool.
            description: str | None = None
                The description of the tool.
            prepare: _pydantic_ai_tools.ToolPrepareFunc[_pydantic_ai_tools.ToolAgentDepsT] | None = None
                The prepare function for the tool.
            docstring_format: _pydantic_ai_tools.DocstringFormat = 'auto'
                The format of the docstring for the tool.
            require_parameter_descriptions: bool = False
                Whether to require parameter descriptions for the tool.
            schema_generator: type[_pydantic_ai_tools.GenerateJsonSchema] = _pydantic_ai_tools.GenerateToolJsonSchema
                The schema generator for the tool.
            strict: bool | None = None
                Whether to enforce strict JSON schema validation for the tool.
            sequential: bool = False
                Whether the tool requires a sequential/serial execution environment.
            requires_approval: bool = False
                Whether the tool requires human-in-the-loop approval.
            metadata: dict[str, Any] | None = None
                The metadata for the tool.
            timeout: float | None = None
                The timeout for the tool.
            function_schema: _function_schema.FunctionSchema | None = None
                The `pydantic_ai._function_schema.FunctionSchema` for the tool.
        """
        super().__init__(
            function=function,
            takes_ctx=takes_ctx,
            max_retries=max_retries,
            name=name,
            description=description,
            prepare=cast(
                _pydantic_ai_tools.ToolPrepareFunc[object] | None, prepare
            ),
            docstring_format=docstring_format,
            require_parameter_descriptions=require_parameter_descriptions,
            schema_generator=schema_generator,
            strict=strict,
            sequential=sequential,
            requires_approval=requires_approval,
            metadata=metadata,
            timeout=timeout,
            function_schema=function_schema,
        )

    def _args_dict(
        self, *args: ToolParams.args, **kwargs: ToolParams.kwargs
    ) -> dict[str, Any]:
        """Build and validate args dict from *args, **kwargs using the tool schema."""
        schema = self.function_schema
        if schema.positional_fields:
            args_dict = dict(zip(schema.positional_fields, args))
        else:
            args_dict = {}
        args_dict.update(kwargs)
        return schema.validator.validate_python(args_dict)

    def __call__(
        self, *args: ToolParams.args, **kwargs: ToolParams.kwargs
    ) -> ToolReturn:
        """Invoke the tool synchronously. Uses pydantic_ai invocation when takes_ctx."""
        if self.takes_ctx:
            ctx = get_current_run_context()
            if ctx is None:
                raise RuntimeError(
                    "Tool requires a run context (takes_ctx=True). "
                    "Call from within an agent run, or use a tool that does not take RunContext."
                )
            args_dict = self._args_dict(*args, **kwargs)
            coro = self.function_schema.call(args_dict, ctx)
            try:
                asyncio.get_running_loop()
            except RuntimeError:
                return asyncio.run(coro)
            raise RuntimeError(
                "Cannot call async tool synchronously from within an async context. "
                "Use await tool.acall(...) instead."
            )
        result = self.function(*args, **kwargs)
        if self.function_schema.is_async:
            try:
                asyncio.get_running_loop()
            except RuntimeError:
                return asyncio.run(result)
            raise RuntimeError(
                "Cannot call async tool synchronously from within an async context. "
                "Use await tool.acall(...) instead."
            )
        return result

    async def acall(
        self, *args: ToolParams.args, **kwargs: ToolParams.kwargs
    ) -> ToolReturn:
        """Invoke the tool asynchronously. Uses pydantic_ai invocation when takes_ctx."""
        if self.takes_ctx:
            ctx = get_current_run_context()
            if ctx is None:
                raise RuntimeError(
                    "Tool requires a run context (takes_ctx=True). "
                    "Call from within an agent run, or use a tool that does not take RunContext."
                )
            args_dict = self._args_dict(*args, **kwargs)
            return await self.function_schema.call(args_dict, ctx)
        result = self.function(*args, **kwargs)
        if self.function_schema.is_async:
            return await result
        return result


def tool(
    function: Callable[ToolParams, ToolReturn]
    | Callable[[PydanticAIRunContext[Deps], ToolParams], ToolReturn],
    *,
    takes_ctx: bool | None = None,
    max_retries: int | None = None,
    name: str | None = None,
    description: str | None = None,
    prepare: _pydantic_ai_tools.ToolPrepareFunc[
        _pydantic_ai_tools.ToolAgentDepsT
    ]
    | None = None,
    docstring_format: _pydantic_ai_tools.DocstringFormat = "auto",
    require_parameter_descriptions: bool = False,
    schema_generator: type[
        _pydantic_ai_tools.GenerateJsonSchema
    ] = _pydantic_ai_tools.GenerateToolJsonSchema,
    strict: bool | None = None,
    sequential: bool = False,
    requires_approval: bool = False,
    metadata: dict[str, Any] | None = None,
    timeout: float | None = None,
    function_schema: _pydantic_ai_function_schema.FunctionSchema | None = None,
) -> Tool[Any, ToolParams, ToolReturn]:
    """Wrapper factory function for creating a `Tool` that can be passed to a model or an agent for tool calling and execution during a semantic operation."""
    return Tool(
        function=function,
        takes_ctx=takes_ctx,
        max_retries=max_retries,
        name=name,
        description=description,
        prepare=cast(
            _pydantic_ai_tools.ToolPrepareFunc[object] | None, prepare
        ),
        docstring_format=docstring_format,
        require_parameter_descriptions=require_parameter_descriptions,
        schema_generator=schema_generator,
        strict=strict,
        sequential=sequential,
        requires_approval=requires_approval,
        metadata=metadata,
        timeout=timeout,
        function_schema=function_schema,
    )



"""zyx.operations.make"""

from __future__ import annotations

import random
from typing import List, Literal, TypeVar, overload, TYPE_CHECKING

from .._aliases import (
    PydanticAIInstructions,
    PydanticAIModelSettings,
    PydanticAIUsageLimits,
)
from .._graph import (
    SemanticGraph,
    SemanticGraphDeps,
    SemanticGraphState,
    SemanticGraphRequestTemplate,
    GraphHooks,
    make_generate_step,
    make_stream_step,
)
from ..result import Result
from ..stream import Stream
from .._types import (
    ModelParam,
    TargetParam,
    ContextType,
    ToolType,
    AttachmentType,
)
from .._utils._semantic import semantic_for_operation

if TYPE_CHECKING:
    from .._utils._observer import Observer

__all__ = (
    "amake",
    "make",
)


Deps = TypeVar("Deps")
Output = TypeVar("Output")


_RANDOMIZATION_KEYWORDS: list[str] = [
    "concise",
    "succinct",
    "evocative",
    "original",
    "minimal",
    "minimalist",
    "uncommon",
    "unusual",
    "abstract",
    "abstractive",
    "poetic",
    "lyrical",
    "philosophical",
    "deep",
    "thoughtful",
    "profound",
    "inspiring",
    "motivational",
    "educational",
    "enlightening",
    "explanatory",
    "analytical",
    "critical",
    "evaluative",
    "judgemental",
    "biased",
    "subjective",
    "opinionated",
    "creative",
    "imaginative",
    "curious",
    "insightful",
    "distinctive",
    "descriptive",
    "innovative",
    "strange",
    "objective",
    "reflective",
]


def _attach_observer_hooks(
    deps: SemanticGraphDeps[Deps, Output], operation: str
) -> None:
    observe = getattr(deps, "observe", None)
    if not observe or getattr(deps, "hooks", None) is not None:
        return
    deps.hooks = GraphHooks(
        on_run_start=lambda _ctx: observe.on_operation_start(operation),
        on_run_end=lambda _res: observe.on_operation_complete(operation),
    )


def prepare_make_graph(
    deps: SemanticGraphDeps[Deps, Output],
    state: SemanticGraphState[Output],
    randomize: bool = False,
    stream: bool = False,
) -> SemanticGraph[Output]:
    """
    Prepares a `SemanticGraph` Graph for the `make` operation, depending on
    it's set configuration.
    """

    randomization_system_prompt = (
        f"Ensure your final response is {random.choice(_RANDOMIZATION_KEYWORDS)} and {random.choice(_RANDOMIZATION_KEYWORDS)}."
        "Although you are free to be creative, try to be as realistic as possible to the given response schema."
        "Your goal is uncommon responses that are believable. Never use the first 2-3 ideas that come to mind."
    )

    if state.output.normalized is str:
        if not deps.message_history and not deps.instructions:
            raise ValueError(
                "Using `make()` with `target=str` requires providing either `context` or `instructions`. Please provide"
                "one of these parameters."
            )

    if not deps.message_history and not deps.instructions:
        request = SemanticGraphRequestTemplate(
            system_prompt_additions=randomization_system_prompt,
            native_output=False if deps.toolsets else True,
        )
        if not deps.agent.model_settings:
            deps.agent.model_settings = {"temperature": 0.7}
    else:
        request = SemanticGraphRequestTemplate(
            system_prompt_additions=randomization_system_prompt
            if randomize
            else None,
            native_output=False if deps.toolsets else True,
        )
        if randomize and not deps.agent.model_settings:
            deps.agent.model_settings = {"temperature": 0.7}

    if stream:
        steps = [make_stream_step(request)]
    else:
        steps = [make_generate_step(request)]
    return SemanticGraph(steps=steps, state=state, deps=deps)


@overload
async def amake(
    target: TargetParam[Output] = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    randomize: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> Stream[Output]: ...


@overload
async def amake(
    target: TargetParam[Output] = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    randomize: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[Output]: ...


async def amake(
    target: TargetParam[Output] = str,  # type: ignore[assignment]
    context: ContextType | List[ContextType] | None = None,
    *,
    randomize: bool = False,
    confidence: bool = False,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    attachments: AttachmentType | List[AttachmentType] | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    observe: bool | Observer | None = None,
    stream: bool = False,
) -> Result[Output] | Stream[Output]:
    """Asynchronously generate ('make') a value of the provided `target` type using
    a model or Pydantic AI agent.

    Args:
        target (TargetParam[Output]): The target type or value to generate. Defaults to str.
        context (ContextType | List[ContextType] | None): The context to use for the operation. Defaults to None.
        randomize (bool): Injects a simple randomization instruction for more diverse outputs. This is automatically
            added if no context or instructions are provided. Defaults to False.
        confidence (bool): Whether to include confidence scores in the result of the operation. This is currently only
            supported for OpenAI or OpenAI-like models. Defaults to False.
        model (ModelParam): The model to use for the operation. This can be a string, Pydantic AI model,
            or Pydantic AI agent. Defaults to "openai:gpt-4o-mini".
        model_settings (PydanticAIModelSettings | None): The model settings to use for the operation. Defaults to None.
        attachments (AttachmentType | List[AttachmentType] | None): A single or list of attachment objects provided to the agent.
            An attachment is a piece of content that is provided to the agent in a 'persistent' fashion,
            where it is templated/placed specifically to avoid context rot or loss. Furthermore, attachments that
            are `Resources` provide the agent with an ability to interact with/modify them, like artifacts. Defaults to None.
        instructions (PydanticAIInstructions | None): The instructions to use for the operation. Defaults to None.
        tools (ToolType | List[ToolType] | None): The tools to use for the operation. Defaults to None.
        deps (Deps | None): Reference to `deps` in `pydantic_ai.RunContext`, that can be passed to messages,
            tools and instructions. Defaults to None.
        usage_limits (PydanticAIUsageLimits | None): The usage limits to use for the operation. Defaults to None.
        observe (bool | Observer | None): If True or provided, enables CLI observation output.
        stream (bool): Whether to stream the output of the operation. Defaults to False.

    Returns:
        Result[Output] | Stream[Output]: The result or stream of the operation.
    """
    graph_deps = SemanticGraphDeps.prepare(
        model=model,
        model_settings=model_settings,
        context=context,
        instructions=instructions,  # type: ignore[arg-type]
        tools=tools,
        confidence=confidence,
        target=target,
        deps=deps,
        usage_limits=usage_limits,
        attachments=attachments,
        observe=observe,
        semantic_renderer=lambda res, _state, _deps: semantic_for_operation(
            "make", output=res.output
        ),
    )
    _attach_observer_hooks(graph_deps, "make")
    state = SemanticGraphState.prepare(deps=graph_deps)

    graph = prepare_make_graph(
        deps=graph_deps, state=state, randomize=randomize, stream=stream
    )

    if stream:
        return await graph.stream()
    else:
        return await graph.run()


@overload
def make(
    target: TargetParam[Output] = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    randomize: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> Stream[Output]: ...


@overload
def make(
    target: TargetParam[Output] = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    randomize: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[Output]: ...


def make(
    target: TargetParam[Output] = str,  # type: ignore[assignment]
    context: ContextType | List[ContextType] | None = None,
    *,
    randomize: bool = False,
    confidence: bool = False,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    attachments: AttachmentType | List[AttachmentType] | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    observe: bool | Observer | None = None,
    stream: bool = False,
) -> Result[Output] | Stream[Output]:
    """Generate ('make') a value of the provided `target` type using
    a model or Pydantic AI agent.

    Args:
        target (TargetParam[Output]): The target type or value to generate. Defaults to str.
        context (ContextType | List[ContextType] | None): The context to use for the operation. Defaults to None.
        randomize (bool): Injects a simple randomization instruction for more diverse outputs. This is automatically
            added if no context or instructions are provided. Defaults to False.
        confidence (bool): Whether to include confidence scores in the result of the operation. This is currently only
            supported for OpenAI or OpenAI-like models. Defaults to False.
        model (ModelParam): The model to use for the operation. This can be a string, Pydantic AI model,
            or Pydantic AI agent. Defaults to "openai:gpt-4o-mini".
        model_settings (PydanticAIModelSettings | None): The model settings to use for the operation. Defaults to None.
        attachments (AttachmentType | List[AttachmentType] | None): A single or list of attachment objects provided to the agent.
            An attachment is a piece of content that is provided to the agent in a 'persistent' fashion,
            where it is templated/placed specifically to avoid context rot or loss. Furthermore, attachments that
            are `Resources` provide the agent with an ability to interact with/modify them, like artifacts. Defaults to None.
        instructions (PydanticAIInstructions | None): The instructions to use for the operation. Defaults to None.
        tools (ToolType | List[ToolType] | None): The tools to use for the operation. Defaults to None.
        deps (Deps | None): Reference to `deps` in `pydantic_ai.RunContext`, that can be passed to messages,
            tools and instructions. Defaults to None.
        usage_limits (PydanticAIUsageLimits | None): The usage limits to use for the operation. Defaults to None.
        observe (bool | Observer | None): If True or provided, enables CLI observation output.
        stream (bool): Whether to stream the output of the operation. Defaults to False.

    Returns:
        Result[Output] | Stream[Output]: The result or stream of the operation.
    """
    graph_deps = SemanticGraphDeps.prepare(
        model=model,
        model_settings=model_settings,
        context=context,
        instructions=instructions,  # type: ignore[arg-type]
        tools=tools,
        confidence=confidence,
        target=target,
        deps=deps,
        usage_limits=usage_limits,
        attachments=attachments,
        observe=observe,
        semantic_renderer=lambda res, _state, _deps: semantic_for_operation(
            "make", output=res.output
        ),
    )
    state = SemanticGraphState.prepare(deps=graph_deps)

    graph = prepare_make_graph(
        deps=graph_deps, state=state, randomize=randomize, stream=stream
    )

    if stream:
        return graph.stream_sync()
    else:
        return graph.run_sync()



"""zyx.processing.messages"""

from __future__ import annotations

from dataclasses import dataclass
import re
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Pattern,
)

from pydantic import TypeAdapter
from pydantic_ai import RunContext
from pydantic_ai.messages import (
    ModelMessage,
    ModelRequest,
    ModelResponse,
    ModelResponsePart,
    SystemPromptPart,
    ToolCallPart,
    ToolReturnPart,
    ToolReturnContent,
    TextPart,
    UserContent,
    UserPromptPart,
    BinaryContent,
    ImageUrl,
)

__all__ = (
    "PydanticAIMessageAdapter",
    "openai_dict_to_pydantic_ai_message",
    "text_to_pydantic_ai_messages",
    "compact_pydantic_ai_messages",
    "parse_context_to_pydantic_ai_messages",
    "parse_instructions_as_system_prompt_parts",
)


PydanticAIMessageAdapter: TypeAdapter[ModelMessage] = TypeAdapter(ModelMessage)


_STRING_MESSAGE_ROLE_PATTERN: Pattern[str] = re.compile(
    pattern=r"\[(s|system|u|user|a|assistant)\](.*?)\[/\1\]",
    flags=re.DOTALL | re.IGNORECASE,
)


_STRING_MESSAGE_UNCLOSED_ROLE_PATTERN: Pattern[str] = re.compile(
    pattern=r"\[(s|system|u|user|a|assistant)\](.*)$",
    flags=re.DOTALL | re.IGNORECASE,
)


_STRING_MESSAGE_ROLE_MAP: Dict[str, str] = {
    "s": "system",
    "system": "system",
    "u": "user",
    "user": "user",
    "a": "assistant",
    "assistant": "assistant",
}


def openai_dict_to_pydantic_ai_message(
    message: Dict[str, Any],
) -> ModelMessage:
    """
    Converts an OpenAI-like message dictionary to a PydanticAI ModelMessage object.
    """
    if "role" not in message or "content" not in message:
        raise ValueError(
            (
                "Recieved invalid dictionary representation of a message. An OpenAI-like message "
                "must contain both a `role` and `content` key to be considered valid."
            )
        )

    role: str = message["role"].lower()
    content: str | List[Dict[str, Any]] = message["content"]

    if role == "system":
        if isinstance(content, str):
            return ModelRequest(parts=[SystemPromptPart(content=content)])
        elif isinstance(content, list):
            parts: List[SystemPromptPart] = []
            for part in content:
                if "type" not in part:
                    raise ValueError(
                        "Found invalid content part within system message content. ",
                        "Each part must contain a `type` key to be considered valid.",
                    )
                if part.get("type") == "text":
                    parts.append(SystemPromptPart(content=part["text"]))

            return ModelRequest(parts=parts)

    elif role == "user":
        if isinstance(content, str):
            return ModelRequest(parts=[UserPromptPart(content=content)])

        elif isinstance(content, list):
            parts: List[UserContent] = []
            for part in content:
                if "type" not in part:
                    raise ValueError(
                        "Found invalid content part within user message content. ",
                        "Each part must contain a `type` key to be considered valid.",
                    )

                if part["type"] == "text":
                    parts.append(part["text"])

                elif part["type"] == "image_url":
                    parts.append(
                        ImageUrl(
                            url=part["image_url"]["url"],
                        )
                    )

                elif part["type"] == "input_audio":
                    parts.append(
                        BinaryContent(
                            data=part["input_audio"]["data"],
                            media_type=part["input_audio"]["format"],
                        )
                    )

                elif part["type"] == "file":
                    parts.append(
                        BinaryContent(
                            data=part["file"]["file_data"],
                            media_type=part["file"]["file_id"],
                        )
                    )

            return ModelRequest(parts=[UserPromptPart(content=parts)])

    elif role == "assistant":
        parsed_content: List[ModelResponsePart] = []

        if "tool_calls" in message and isinstance(message["tool_calls"], list):
            # assume tool calls are in correct format, if theyre being passed in
            # as a dict. no need to validate every little bit
            for tool_call in message["tool_calls"]:
                parsed_content.append(
                    ToolCallPart(
                        tool_name=tool_call["function"]["name"],
                        args=tool_call["function"]["arguments"],
                        tool_call_id=tool_call["id"],
                    )
                )

        if content is not None:
            if isinstance(content, str):
                parsed_content.append(TextPart(content=content))

            elif isinstance(content, list):
                for part in content:
                    if part.get("type") == "text":
                        parsed_content.append(TextPart(content=part["text"]))

                # TODO: decide if theres a need to support additional content parts
                # at this level of dictionary parsing

        return ModelResponse(parts=parsed_content)

    elif role == "tool":
        if isinstance(content, str):
            return ModelRequest(
                parts=[
                    ToolReturnPart(tool_name=message["name"], content=content)
                ]
            )

        elif isinstance(content, list):
            parts: List[ToolReturnContent] = []
            for part in content:
                if part.get("type") == "text":
                    parts.append(part["text"])

            return ModelRequest(
                parts=[
                    ToolReturnPart(tool_name=message["name"], content=parts)
                ]
            )

    raise ValueError(
        "Recieved invalid dictionary representation of an OpenAI ChatCompletions message. ",
        f"A message must match one of the following roles: 'user', 'system', 'assistant', 'tool', received: '{role}'",
    )


def text_to_pydantic_ai_messages(text: str) -> List[ModelMessage]:
    """
    Converts a string to a list of PydanticAI ModelMessage objects.

    This supports the use of 'role tags' to include multiple messages within
    a single string. Supported tags are:

    - [s] or [system] for system messages
    - [u] or [user] for user messages
    - [a] or [assistant] for assistant messages

    Ex : [s] You are a helpful assistant.[/s] [u] Hello, how are you?[/u]
    """
    messages: List[ModelMessage] = []
    # index of the last end tagged section
    last_end: int = 0

    for match in _STRING_MESSAGE_ROLE_PATTERN.finditer(text):
        # if any text is untagged, and matches are present, this is considered
        # a user message
        if match.start() > last_end:
            untagged = text[last_end : match.start()].strip()
            if untagged:
                messages.append(ModelRequest(parts=[UserPromptPart(untagged)]))

        tag = match.group(1).lower()
        content = match.group(2).strip()
        role = _STRING_MESSAGE_ROLE_MAP[tag]

        if content:
            if role == "system":
                messages.append(
                    ModelRequest(parts=[SystemPromptPart(content=content)])
                )
            elif role == "user":
                messages.append(
                    ModelRequest(parts=[UserPromptPart(content=content)])
                )
            elif role == "assistant":
                messages.append(
                    ModelResponse(parts=[TextPart(content=content)])
                )

            # NOTE:
            # any other tags that arent system, user, or assistant are ignored, and
            # the text is considered untagged
            else:
                messages.append(
                    ModelRequest(parts=[UserPromptPart(content=content)])
                )

        last_end = match.end()

    if last_end < len(text):
        # if additional content/text is present, beyond any tags, check if it's
        # an unclosed tag at the end
        remaining = text[last_end:].strip()
        if remaining:
            # check if remaining text starts with an unclosed tag
            unclosed_match = _STRING_MESSAGE_UNCLOSED_ROLE_PATTERN.match(
                remaining
            )
            if unclosed_match:
                tag = unclosed_match.group(1).lower()
                content = unclosed_match.group(2).strip()
                role = _STRING_MESSAGE_ROLE_MAP[tag]

                if content:
                    if role == "system":
                        messages.append(
                            ModelRequest(
                                parts=[SystemPromptPart(content=content)]
                            )
                        )
                    elif role == "user":
                        messages.append(
                            ModelRequest(
                                parts=[UserPromptPart(content=content)]
                            )
                        )
                    elif role == "assistant":
                        messages.append(
                            ModelResponse(parts=[TextPart(content=content)])
                        )
                    else:
                        # fallback to user message for unknown roles
                        messages.append(
                            ModelRequest(
                                parts=[UserPromptPart(content=content)]
                            )
                        )
            else:
                # no unclosed tag, treat as user message
                messages.append(
                    ModelRequest(parts=[UserPromptPart(remaining)])
                )

    # finally, once again we fallback to a user message
    if not messages and text.strip():
        messages.append(ModelRequest(parts=[UserPromptPart(text.strip())]))

    return messages


def compact_pydantic_ai_messages(
    messages: List[ModelMessage],
) -> List[ModelMessage]:
    """
    Compacts consecutive ModelRequest or ModelResponse messages into single messages, if applicable.
    """
    if not messages:
        return messages

    compacted: List[ModelMessage] = []

    for message in messages:
        if not compacted:
            compacted.append(message)
            continue

        last_message = compacted[-1]

        if isinstance(message, ModelRequest) and isinstance(
            last_message, ModelRequest
        ):
            compacted[-1] = ModelRequest(
                parts=list(last_message.parts) + list(message.parts)
            )
        elif isinstance(message, ModelResponse) and isinstance(
            last_message, ModelResponse
        ):
            compacted[-1] = ModelResponse(
                parts=list(last_message.parts) + list(message.parts)
            )
        else:
            compacted.append(message)

    return compacted


def parse_context_to_pydantic_ai_messages(
    context: Any | List[Any] | None = None,
) -> List[ModelMessage]:
    """
    Prepares the `context` parameter of a semantic operation into a list of
    `ModelMessage` objects.
    """
    messages: List[ModelMessage] = []

    from ..context import Context
    from ..attachments import Attachment

    if not context:
        return []
    else:
        if not isinstance(context, list):
            context = [context]

        for item in context:
            if isinstance(item, Context):
                messages.extend(item.render_messages())
                continue

            if isinstance(item, str):
                messages.extend(text_to_pydantic_ai_messages(item))
                continue

            if isinstance(item, Attachment):
                msg = item.message
                if msg is not None:
                    messages.append(msg)
                continue

            if hasattr(item, "model_dump"):
                item = item.model_dump()

            if isinstance(item, dict):
                if "parts" in item:
                    try:
                        messages.append(
                            PydanticAIMessageAdapter.validate_python(item)
                        )
                    except Exception as e:
                        raise ValueError(
                            f"Failed to convert PydanticAI-like message dictionary to a ModelMessage object: {e}"
                        )
                elif "role" in item:
                    try:
                        messages.append(
                            openai_dict_to_pydantic_ai_message(item)
                        )
                    except Exception as e:
                        raise ValueError(
                            f"Failed to convert OpenAI-like message dictionary to a ModelMessage object: {e}"
                        )
                else:
                    raise ValueError(f"Invalid message item: {item}")
        return compact_pydantic_ai_messages(messages)


@dataclass
class _RunContext:
    """
    Internal class used to pass `deps` to functions that represent instructions.
    """

    deps: Any | None = None


def parse_instructions_as_system_prompt_parts(
    instructions: str | Callable | List[str | Callable],
    deps: Any | None = None,
    compact: bool = False,
) -> List[SystemPromptPart]:
    """
    Parses a given `instructions` parameter, that can contain a single or list
    of strings, or functions that optionally accept `pydantic_ai.RunContext[Deps]`
    dependencies.
    """
    if not instructions:
        return []

    if isinstance(instructions, SystemPromptPart):
        return [instructions]
    if not isinstance(instructions, list):
        instructions = [instructions]

    parts: List[str] = []

    for instruction in instructions:
        if inspect.isfunction(instruction):
            signature = inspect.signature(instruction)
            params = signature.parameters
            should_pass_deps = False

            if params:
                first_param = next(iter(params.values()))
                if deps is not None:
                    if first_param.annotation is not inspect._empty and (
                        first_param.annotation == RunContext
                        or (
                            hasattr(first_param.annotation, "__origin__")
                            and first_param.annotation.__origin__ == RunContext
                        )
                    ):
                        should_pass_deps = True
                    elif len(params) == 1:
                        should_pass_deps = True
                elif len(params) == 1:
                    should_pass_deps = True

            try:
                resolved = instruction(
                    _RunContext(deps=deps) if should_pass_deps else None
                )
            except Exception:
                resolved = instruction(
                    _RunContext(deps=deps) if deps is not None else None
                )

            if not resolved:
                pass
            else:
                parts.append(str(resolved))

        elif isinstance(instruction, str):
            parts.append(instruction)
        else:
            raise ValueError(f"Invalid instruction: {instruction}")

    if compact:
        prompt = "\n\n".join(parts)
        return [SystemPromptPart(content=prompt)]
    else:
        return [SystemPromptPart(content=part) for part in parts]



"""zyx.operations.select"""

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Iterable,
    List,
    Literal,
    Sequence,
    TypeVar,
    overload,
    cast,
    TYPE_CHECKING,
)

from pydantic import BaseModel

from .._aliases import (
    PydanticAIInstructions,
    PydanticAIModelSettings,
    PydanticAIUsageLimits,
)
from .._graph import (
    SemanticGraph,
    SemanticGraphDeps,
    SemanticGraphState,
    SemanticGraphRequestTemplate,
    GraphHooks,
    make_generate_step,
    make_stream_step,
)
from .._processing._outputs import selection_output_model
from .._types import (
    ContextType,
    ModelParam,
    ToolType,
)
from .._utils._semantic import semantic_for_operation
from ..result import Result
from ..stream import Stream

if TYPE_CHECKING:
    from .._utils._observer import Observer

__all__ = ("aselect", "select")


Deps = TypeVar("Deps")
Output = TypeVar("Output")


class SelectionWithReason(BaseModel):
    selection: Any
    reason: str | None = None


class MultiSelectionWithReason(BaseModel):
    selection: list[Any]
    reason: str | None = None


_SELECT_SYSTEM_PROMPT = (
    "\n[INSTRUCTION]\n"
    "- Your ONLY task is to select the best option(s) from the provided choices.\n"
    "- You MUST respond with a value that matches the given structured output schema "
    "exactly (no explanations, no conversational text, no extra keys).\n"
    "- The available choices are described in the schema docstring. Use them as the "
    "ground truth set of options.\n"
    "- Do NOT invent new options or modify the option values.\n"
    "- For single selection, choose exactly one option. For multi-selection, choose "
    "zero or more options.\n"
    "- Never return the literal value null as the entire response.\n"
)


def _attach_observer_hooks(
    deps: SemanticGraphDeps[Deps, BaseModel], operation: str
) -> None:
    observe = getattr(deps, "observe", None)
    if not observe or getattr(deps, "hooks", None) is not None:
        return
    deps.hooks = GraphHooks(
        on_run_start=lambda _ctx: observe.on_operation_start(operation),
        on_run_end=lambda _res: observe.on_operation_complete(operation),
    )


def _normalize_choices(target: Any | Sequence[Any]) -> list[Any]:
    """Normalize the user-provided `target` into a flat list of concrete choices.

    This mirrors the behaviour expected by `selection_output_model` and supports:
    - list[...] of values or types
    - Literal[...]
    - Enum subclasses
    - Union[..., ...]
    """
    from enum import Enum
    from typing import get_args, get_origin

    # Already a concrete list of choices
    if isinstance(target, list):
        return list(target)

    origin = get_origin(target)
    args = get_args(target)

    # Enum class
    if isinstance(target, type) and issubclass(target, Enum):
        return list(target)

    # Literal[...] or Union[...] or list[T]
    if origin is not None:
        # list[T] – treat each element type/value as a choice
        if origin is list:
            return list(args)
        # Literal[...] / Union[...]
        return list(args)

    # Fallback: single value/type treated as the only choice
    return [target]


def _selection_to_output(
    target: Any | Sequence[Any],
    output: BaseModel,
    *,
    multi_label: bool = False,
    literal: bool = False,
    include_reason: bool = False,
) -> Any:
    """Map a structured selection model back to the original target object(s).

    The return value is:
    - A single selected object when `multi_label` is False
    - A list of selected objects when `multi_label` is True
    """

    # Try to re-use the choices stored on the model built by `selection_output_model`
    model_cls = type(output)
    choices: list[Any] | None = getattr(model_cls, "_choices", None)
    if choices is None:
        choices = _normalize_choices(target)

    def _select_one(idx_or_value: Any) -> Any:
        # Integer index into the choices list
        if isinstance(idx_or_value, int):
            if not choices:
                raise ValueError("No choices available for selection.")
            # Reject out-of-range indices (avoid accidental clamping)
            if idx_or_value < 0 or idx_or_value >= len(choices):
                return None
            return choices[idx_or_value]

        # Literal mode or direct value – match by equality against choices
        for choice in choices:
            if choice == idx_or_value:
                return choice
        # If nothing matches, return the raw value as-is
        return idx_or_value

    if multi_label:
        indices: Iterable[Any] | None = getattr(output, "indices", None)
        if indices is None:
            # Graceful fallback – treat missing indices as empty selection
            selection: Any = []
        else:
            selected: list[Any] = []
            for value in indices:
                mapped = _select_one(value)
                if mapped is not None:
                    selected.append(mapped)
            selection = selected
    else:
        index_value: Any = getattr(output, "index", None)
        if index_value is None:
            # No explicit selection – return None
            selection = None
        else:
            selection = _select_one(index_value)

    if include_reason:
        reason: Any = getattr(output, "reason", None)
        if multi_label:
            return MultiSelectionWithReason(
                selection=cast(list[Any], selection), reason=reason
            )
        return SelectionWithReason(selection=selection, reason=reason)

    return selection


def prepare_select_graph(
    deps: SemanticGraphDeps[Deps, BaseModel],
    state: SemanticGraphState[BaseModel],
    *,
    multi_label: bool = False,
    literal: bool = False,
    include_reason: bool = False,
    stream: bool = False,
) -> SemanticGraph[BaseModel]:
    """
    Prepare a `SemanticGraph` for the `select` operation.

    The graph always generates an intermediate structured selection model; the
    public `select` / `aselect` helpers then map this back to the actual
    selected object(s).
    """
    # The deps.target has already been set to the selection model type.
    selection_model_type = cast(type[BaseModel], deps.target)

    if not deps.confidence:
        request = SemanticGraphRequestTemplate(
            output_type=selection_model_type,
            system_prompt_additions=_SELECT_SYSTEM_PROMPT,
            native_output=False if deps.toolsets else True,
        )
    else:
        request = SemanticGraphRequestTemplate(
            output_type=selection_model_type,
            system_prompt_additions=_SELECT_SYSTEM_PROMPT,
            native_output=True,
        )

    if stream:
        steps = [make_stream_step(request)]
    else:
        steps = [make_generate_step(request)]

    return SemanticGraph(steps=steps, state=state, deps=deps)


@dataclass
class _SelectionStreamWrapper:
    """Lightweight wrapper that maps the final streamed selection model back
    to the actual selected object(s), while preserving the underlying stream
    behaviour for text/partial/field.
    """

    _inner: Stream[BaseModel]
    _target: Any | Sequence[Any]
    _multi_label: bool
    _literal: bool
    _include_reason: bool

    async def __aenter__(self) -> "_SelectionStreamWrapper":
        await self._inner.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self._inner.__aexit__(exc_type, exc_val, exc_tb)

    # --- Streaming primitives delegate to the inner stream unchanged ---

    def text(self, *, delta: bool = False, debounce_by: float | None = 0.1):
        return self._inner.text(delta=delta, debounce_by=debounce_by)

    def partial(self, *, debounce_by: float | None = 0.1):
        return self._inner.partial(debounce_by=debounce_by)

    def field(self, field_name: str, *, debounce_by: float | None = 0.1):
        return self._inner.field(
            field_name=field_name, debounce_by=debounce_by
        )

    @property
    def is_streaming(self) -> bool:
        return self._inner.is_streaming

    @property
    def usage(self):
        return self._inner.usage

    # --- Finalisation helpers that remap the output ---

    async def finish_async(self) -> Result[Output]:
        base_result = await self._inner.finish_async()
        selected = _selection_to_output(
            self._target,
            base_result.output,
            multi_label=self._multi_label,
            literal=self._literal,
            include_reason=self._include_reason,
        )
        return Result(output=selected, raw=base_result.raw)

    def finish(self) -> Result[Output]:
        # Synchronous wrapper around `finish_async`, mirroring `Stream.finish`.
        import asyncio

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return loop.run_until_complete(self.finish_async())


@overload
async def aselect(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: Literal[False] = ...,
    literal: bool = ...,
    include_reason: Literal[True],
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> _SelectionStreamWrapper: ...


@overload
async def aselect(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: bool = ...,
    literal: bool = ...,
    include_reason: Literal[True],
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[SelectionWithReason]: ...


@overload
async def aselect(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: Literal[True],
    literal: bool = ...,
    include_reason: Literal[True],
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[MultiSelectionWithReason]: ...


@overload
async def aselect(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: bool = ...,
    literal: bool = ...,
    include_reason: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> _SelectionStreamWrapper: ...


@overload
async def aselect(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: bool = ...,
    literal: bool = ...,
    include_reason: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[Output]: ...


async def aselect(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = None,
    *,
    multi_label: bool = False,
    literal: bool = True,
    include_reason: bool = False,
    confidence: bool = False,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    observe: bool | Observer | None = None,
    stream: bool = False,
) -> Result[Output] | _SelectionStreamWrapper:
    """Asynchronously select one or more options from the given `target`.

    Args:
        target (Any | List[Any]): The options to select from. Can be:
            - A list of values or types
            - A Literal[...] type
            - An Enum subclass
            - A Union[..., ...] type
        context (ContextType | List[ContextType] | None): Optional additional context or conversation history. Defaults to None.
        multi_label (bool): When True, return a list of selected objects; otherwise
            return a single selected object. Defaults to False.
        literal (bool): When True and supported by the options, use Literal[...] for
            the selection field instead of indices. Defaults to True.
        include_reason (bool): Reserved flag to include a textual reason field in the
            intermediate structured output (not exposed in the final result). Defaults to False.
        confidence (bool): When True, enable log-probability based confidence scoring. Defaults to False.
        model (ModelParam): The model to use for selection. Defaults to "openai:gpt-4o-mini".
        model_settings (PydanticAIModelSettings | None): Model settings to pass to the operation (e.g., temperature). Defaults to None.
        instructions (PydanticAIInstructions | None): Additional instructions/hints for the model. Defaults to None.
        tools (ToolType | List[ToolType] | None): List of tools available to the model. Defaults to None.
        deps (Deps | None): Optional dependencies (e.g., `pydantic_ai.RunContext`) for this operation. Defaults to None.
        usage_limits (PydanticAIUsageLimits | None): Usage limits (token/request) configuration. Defaults to None.
        stream (bool): When True, return a stream wrapper that exposes the underlying
            model stream and remaps the final result to the selected object(s). Defaults to False.

    Returns:
        Result[Output] | _SelectionStreamWrapper: A `Result` whose `output` is either the selected object or a list
            of selected objects. When `include_reason` is True, the output is a
            dict with keys `selection` and `reason`. For streaming, the wrapper
            preserves the same final semantics.
    """
    # Build the structured selection model for these options.
    selection_model = selection_output_model(
        options=target,
        name="Selection",
        multi_label=multi_label,
        literal=literal,
        reason=include_reason,
    )

    graph_deps = SemanticGraphDeps.prepare(
        model=model,
        model_settings=model_settings,
        context=context,
        instructions=instructions,  # type: ignore[arg-type]
        tools=tools,
        deps=deps,
        usage_limits=usage_limits,
        target=selection_model,
        source=target,
        confidence=confidence,
        observe=observe,
        semantic_renderer=lambda res, _state, _deps: semantic_for_operation(
            "select", output=res.output
        ),
    )
    _attach_observer_hooks(graph_deps, "select")
    graph_state = SemanticGraphState.prepare(deps=graph_deps)

    graph = prepare_select_graph(
        deps=graph_deps,
        state=graph_state,
        multi_label=multi_label,
        literal=literal,
        include_reason=include_reason,
        stream=stream,
    )

    if stream:
        inner_stream: Stream[BaseModel] = await graph.stream()
        return _SelectionStreamWrapper(
            _inner=inner_stream,
            _target=target,
            _multi_label=multi_label,
            _literal=literal,
            _include_reason=include_reason,
        )

    base_result = await graph.run()
    selected = _selection_to_output(
        target,
        base_result.output,
        multi_label=multi_label,
        literal=literal,
        include_reason=include_reason,
    )
    return Result(output=selected, raw=base_result.raw)


@overload
def select(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: Literal[False] = ...,
    literal: bool = ...,
    include_reason: Literal[True],
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> _SelectionStreamWrapper: ...


@overload
def select(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: bool = ...,
    literal: bool = ...,
    include_reason: Literal[True],
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[SelectionWithReason]: ...


@overload
def select(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: Literal[True],
    literal: bool = ...,
    include_reason: Literal[True],
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[MultiSelectionWithReason]: ...


@overload
def select(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: bool = ...,
    literal: bool = ...,
    include_reason: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> _SelectionStreamWrapper: ...


@overload
def select(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = ...,
    *,
    multi_label: bool = ...,
    literal: bool = ...,
    include_reason: bool = ...,
    confidence: bool = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> Result[Output]: ...


def select(
    target: Any | List[Any],
    context: ContextType | List[ContextType] | None = None,
    *,
    multi_label: bool = False,
    literal: bool = True,
    include_reason: bool = False,
    confidence: bool = False,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    observe: bool | Observer | None = None,
    stream: bool = False,
) -> Result[Output] | _SelectionStreamWrapper:
    """Synchronously select one or more options from the given `target`.

    Args:
        target (Any | List[Any]): The options to select from. Can be:
            - A list of values or types
            - A Literal[...] type
            - An Enum subclass
            - A Union[..., ...] type
        context (ContextType | List[ContextType] | None): Optional additional context or conversation history. Defaults to None.
        multi_label (bool): When True, return a list of selected objects; otherwise
            return a single selected object. Defaults to False.
        literal (bool): When True and supported by the options, use Literal[...] for
            the selection field instead of indices. Defaults to True.
        include_reason (bool): Reserved flag to include a textual reason field in the
            intermediate structured output (not exposed in the final result). Defaults to False.
        confidence (bool): When True, enable log-probability based confidence scoring. Defaults to False.
        model (ModelParam): The model to use for selection. Defaults to "openai:gpt-4o-mini".
        model_settings (PydanticAIModelSettings | None): Model settings to pass to the operation (e.g., temperature). Defaults to None.
        instructions (PydanticAIInstructions | None): Additional instructions/hints for the model. Defaults to None.
        tools (ToolType | List[ToolType] | None): List of tools available to the model. Defaults to None.
        deps (Deps | None): Optional dependencies (e.g., `pydantic_ai.RunContext`) for this operation. Defaults to None.
        usage_limits (PydanticAIUsageLimits | None): Usage limits (token/request) configuration. Defaults to None.
        stream (bool): When True, return a stream wrapper that exposes the underlying
            model stream and remaps the final result to the selected object(s). Defaults to False.

    Returns:
        Result[Output] | _SelectionStreamWrapper: A `Result` whose `output` is either the selected object or a list
            of selected objects. When `include_reason` is True, the output is a
            dict with keys `selection` and `reason`. For streaming, the wrapper
            preserves the same final semantics.
    """
    # Build the structured selection model for these options.
    selection_model = selection_output_model(
        options=target,
        name="Selection",
        multi_label=multi_label,
        literal=literal,
        reason=include_reason,
    )

    graph_deps = SemanticGraphDeps.prepare(
        model=model,
        model_settings=model_settings,
        context=context,
        instructions=instructions,  # type: ignore[arg-type]
        tools=tools,
        deps=deps,
        usage_limits=usage_limits,
        target=selection_model,
        source=target,
        confidence=confidence,
        observe=observe,
        semantic_renderer=lambda res, _state, _deps: semantic_for_operation(
            "select", output=res.output
        ),
    )
    _attach_observer_hooks(graph_deps, "select")
    graph_state = SemanticGraphState.prepare(deps=graph_deps)

    graph = prepare_select_graph(
        deps=graph_deps,
        state=graph_state,
        multi_label=multi_label,
        literal=literal,
        include_reason=include_reason,
        stream=stream,
    )

    if stream:
        inner_stream: Stream[BaseModel] = graph.stream_sync()
        return _SelectionStreamWrapper(
            _inner=inner_stream,
            _target=target,
            _multi_label=multi_label,
            _literal=literal,
            _include_reason=include_reason,
        )

    base_result = graph.run_sync()
    selected = _selection_to_output(
        target,
        base_result.output,
        multi_label=multi_label,
        literal=literal,
        include_reason=include_reason,
    )
    return Result(output=selected, raw=base_result.raw)



"""zyx.context"""

from __future__ import annotations

from dataclasses import dataclass, field
import inspect
from typing import Any, Callable, Generic, List, TypeVar, TYPE_CHECKING

from pydantic_ai import messages as _messages
from pydantic_ai.toolsets import FunctionToolset

from ._aliases import (
    PydanticAIInstructions,
    PydanticAIMessage,
    PydanticAIModelRequest,
    PydanticAIModelResponse,
    PydanticAIAgentResult,
    PydanticAIBuiltinTool,
    PydanticAITool,
    PydanticAIToolset,
    PydanticAIUserContent,
)
from ._processing._messages import (
    parse_context_to_pydantic_ai_messages,
    parse_instructions_as_system_prompt_parts,
)

if TYPE_CHECKING:
    from ._types import ToolType, ContextType

__all__ = (
    "Context",
    "create_context",
)


Deps = TypeVar("Deps")


@dataclass(init=False)
class Context(Generic[Deps]):
    """Simple and mutable container for the conversation/message history across
    multiple semantic operations, with the optional ability to set context-specific
    instructions, tools and `deps` that can be passed to `pydantic_ai`'s RunContext."""

    _messages: List[PydanticAIMessage] = field(default_factory=list)

    instructions: str | Callable[[Deps | None], str] | None = None
    """System-level instructions, either a string or a callable. If callable,
    it may optionally accept `deps` and should return a string."""

    tools: List[ToolType] = field(default_factory=list)
    """Tools to include in the context. This can be a sequence of functions, pydantic ai tools, builtin tools
    or toolsets."""

    deps: Deps | None = None
    """`deps` that can be passed to messages, tools and instructions through `pydantic_ai`'s
    RunContext."""

    update: bool = True
    """When `True` (default), semantic operations will automatically update this context after their result
    is completed."""

    compact_instructions: bool = False
    """When `True`, all instructions are compacted into a single `SystemPromptPart`; otherwise they are split on double newlines."""

    exclude_messages: bool = False
    """When `True`, existing messages are not forwarded to operations."""

    exclude_instructions: bool = False
    """When `True`, instructions are not rendered into messages."""

    exclude_tools: bool = False
    """When `True`, tools are not rendered into messages."""

    max_length: int | None = None
    """Optional cap on the number of messages forwarded, from the end of the message history.
    This does not affect the inclusion of instructions."""

    def __init__(
        self,
        messages: ContextType | List[ContextType] | None = None,
        instructions: PydanticAIInstructions | None = None,
        tools: ToolType | List[ToolType] | None = None,
        deps: Deps | None = None,
        update: bool = True,
        compact_instructions: bool = False,
        exclude_messages: bool = False,
        exclude_instructions: bool = False,
        exclude_tools: bool = False,
        max_length: int | None = None,
    ) -> None:
        """
        Initializes a new `Context` object with the given parameters.

        Args:
            messages : Sequence[Any] | None = None
                Initial messages to seed the context. This is any value accepted by
                `parse_context_to_pydantic_ai_messages`.
            instructions : str | Callable[[Deps | None], str] | None = None
                System-level instructions, either a string or a callable. If callable,
                it may optionally accept `deps` and should return a string.
            tools : Sequence[ToolType] | ToolType | None = None
                Tools to include in the context. This can be a sequence of functions, pydantic ai tools, builtin tools
                or toolsets.
            deps : Deps | None = None
                Reference to `deps` in `pydantic_ai.RunContext`, that can be passed to messages,
                tools and instructions.
            update : bool = True
                When `True` (default), semantic operations that consume this context will write back the full message history.
            compact_instructions : bool = False
                When `True`, all instructions are compacted into a single `SystemPromptPart`; otherwise they are split on double newlines.
            exclude_messages : bool = False
                When `True`, existing messages are not forwarded to operations.
            exclude_instructions : bool = False
                When `True`, instructions are not rendered into messages.
            exclude_tools : bool = False
                When `True`, tools are not rendered into messages.
            max_length : int | None = None
                Optional cap on the number of messages forwarded (from the tail).
        """
        if messages:
            self._messages = parse_context_to_pydantic_ai_messages(messages)
        else:
            self._messages = []

        self.instructions = instructions  # type: ignore

        self.tools: List[ToolType] = []

        if tools:
            if not isinstance(tools, list):
                tools = [tools]
            self.tools = tools  # type: ignore
        else:
            self.tools = []

        self.deps = deps
        self.update = update
        self.compact_instructions = compact_instructions
        self.exclude_messages = exclude_messages
        self.exclude_instructions = exclude_instructions
        self.exclude_tools = exclude_tools
        self.max_length = max_length

    def __call__(
        self,
        *,
        instructions: PydanticAIInstructions | None = None,
        tools: ToolType | List[ToolType] | None = None,
        deps: Deps | None = None,
        update: bool | None = None,
        compact_instructions: bool | None = None,
        exclude_messages: bool | None = None,
        exclude_instructions: bool | None = None,
        exclude_tools: bool | None = None,
        max_length: int | None = None,
    ) -> Context[Deps]:
        """Update the content and configuration of this `Context` with per-call overrides and
        return a new `Context` with the updated content.

        NOTE: This method creates a copy of this `Context` object before updates
        are applied, so no updates are set to the original `Context` object.

        Args:
            instructions : PydanticAIInstructions | None = None
                System-level instructions, either a string or a callable. If callable,
                it may optionally accept `deps` and should return a string.
            tools : ToolType | List[ToolType] | None = None
                Tools to include in the context. This can be a sequence of functions, pydantic ai tools, builtin tools
                or toolsets.
            deps : Deps | None = None
                Reference to `deps` in `pydantic_ai.RunContext`, that can be passed to messages,
                tools and instructions.
            update : bool | None = None
                When `True` (default), semantic operations that consume this context will write back the full message history.
            compact_instructions : bool | None = None
                When `True`, all instructions are compacted into a single `SystemPromptPart`; otherwise they are split on double newlines.
            exclude_messages : bool | None = False
                When `True`, existing messages are not forwarded to operations.
            exclude_instructions : bool | None = False
                When `True`, instructions are not rendered into messages.
            exclude_tools : bool | None = False
                When `True`, tools are not rendered into messages.
            max_length : int | None = None
                Optional cap on the number of messages forwarded, from the end of the message history.
                This does not affect the inclusion of instructions.
        """
        cls = self.copy()

        if instructions is not None:
            cls.instructions = instructions  # type: ignore
        if tools is not None:
            if not isinstance(tools, list):
                tools = [tools]
            cls.tools = tools  # type: ignore
        if deps is not None:
            cls.deps = deps
        if update is not None:
            cls.update = update
        if compact_instructions is not None:
            cls.compact_instructions = compact_instructions
        if exclude_messages is not None:
            cls.exclude_messages = exclude_messages
        if exclude_instructions is not None:
            cls.exclude_instructions = exclude_instructions
        if exclude_tools is not None:
            cls.exclude_tools = exclude_tools
        if max_length is not None:
            cls.max_length = max_length

        return cls

    def copy(self) -> Context[Deps]:
        cls = Context(
            messages=None,
            instructions=self.instructions,
            tools=self.tools,
            deps=self.deps,
            update=self.update,
            compact_instructions=self.compact_instructions,
            exclude_messages=self.exclude_messages,
            exclude_instructions=self.exclude_instructions,
            exclude_tools=self.exclude_tools,
            max_length=self.max_length,
        )
        cls._messages = self._messages

        return cls

    def clear(self) -> None:
        """Reset the message history of this `Context` to an empty list."""
        self._messages.clear()

    def add_user_message(
        self,
        content: str | List[PydanticAIUserContent],
        index: int | None = None,
    ) -> None:
        """
        Append or inject a user message to the message history if `index`
        is given.

        Args:
            content : str | List[PydanticAIUserContent]
                The content of the user message. This can be a string or a list of `PydanticAIUserContent` objects.
            index : int | None = None
                The index at which to inject the user message. If not given,
                the message is appended to the end of the message history.
        """
        if index is not None:
            self._messages.insert(
                index,
                PydanticAIModelRequest(
                    parts=[_messages.UserPromptPart(content=content)]
                ),
            )
        else:
            self._messages.append(
                PydanticAIModelRequest(
                    parts=[_messages.UserPromptPart(content=content)]
                )
            )

    def add_assistant_message(
        self,
        content: str,
        index: int | None = None,
    ) -> None:
        """
        Append or inject an assistant message to the message history if `index`
        is given.

        Args:
            content : str
                The content of the assistant message.
            index : int | None = None
                The index at which to inject the assistant message. If not given,
                the message is appended to the end of the message history.
        """
        if index is not None:
            self._messages.insert(
                index,
                PydanticAIModelResponse(
                    parts=[_messages.TextPart(content=content)]
                ),
            )
        else:
            self._messages.append(
                PydanticAIModelResponse(
                    parts=[_messages.TextPart(content=content)]
                )
            )

    def add_system_message(
        self,
        content: str,
        index: int | None = None,
    ) -> None:
        """
        Append or inject a system message to the message history if `index`
        is given.

        Args:
            content : str
                The content of the system message.
            index : int | None = None
                The index at which to inject the system message. If not given,
                the message is appended to the end of the message history.
        """
        if index is not None:
            self._messages.insert(
                index,
                PydanticAIModelRequest(
                    parts=[_messages.SystemPromptPart(content=content)]
                ),
            )
        else:
            self._messages.append(
                PydanticAIModelRequest(
                    parts=[_messages.SystemPromptPart(content=content)]
                )
            )

    def render_messages(self) -> List[PydanticAIMessage]:
        """Renders the message history of this `Context` into a list of `PydanticAIMessage` objects."""
        if self.exclude_messages:
            return []

        messages = list(self._messages)

        if self.max_length is not None and self.max_length >= 0:
            if len(messages) > self.max_length:
                messages = messages[-self.max_length :]

        return messages

    def render_instructions(self) -> List[_messages.SystemPromptPart]:  # type: ignore
        """Render instructions into one or more ``SystemPromptPart`` objects."""
        if self.exclude_instructions or not self.instructions:
            return []

        return parse_instructions_as_system_prompt_parts(
            instructions=self.instructions,
            deps=self.deps,
            compact=self.compact_instructions,
        )

    def render_toolsets(self) -> List[PydanticAIToolset]:
        """Render the toolsets of this `Context` into a list of `PydanticAIToolset` objects."""
        if self.exclude_tools:
            return []

        toolsets = []
        function_tools = []

        for tool in self.tools:
            if inspect.isfunction(tool):
                function_tools.append(PydanticAITool(function=tool))
            elif isinstance(tool, (PydanticAITool, PydanticAIBuiltinTool)):
                function_tools.append(tool)
            elif isinstance(tool, PydanticAIToolset):
                toolsets.append(tool)
            else:
                raise ValueError(f"Invalid tool type: {type(tool)}")

        if function_tools:
            toolsets.append(FunctionToolset(tools=function_tools))

        return toolsets

    def update_from_pydantic_ai_result(
        self, result: PydanticAIAgentResult[Any]
    ) -> None:
        """Update the message history of this `Context` with the full exchange from a run."""
        try:
            self._messages.extend(result.new_messages())
        except Exception:
            pass

    def extend_messages(self, messages: List[PydanticAIMessage]) -> None:
        """Append a list of PydanticAI messages to this context."""
        self._messages.extend(messages)

    def __rich__(self):
        from rich.console import RenderableType, Group
        from rich.rule import Rule
        from rich.text import Text

        renderables: list[RenderableType] = []

        renderables.append(
            Rule(title="✨ Context", style="rule.line", align="left")
        )

        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Messages: {len(self._messages)}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Instructions: {self.instructions!r}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Tools: {len(self.tools)}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Deps: {bool(self.deps)}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Update on Result: {self.update}[/dim italic]"
            )
        )

        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Compact Instructions: {self.compact_instructions}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Exclude Messages: {self.exclude_messages}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Exclude Instructions: {self.exclude_instructions}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Exclude Tools: {self.exclude_tools}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Max Length: {self.max_length}[/dim italic]"
            )
        )

        return Group(*renderables)


def create_context(
    *,
    messages: ContextType | List[ContextType] | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    update: bool = True,
    compact_instructions: bool = False,
    exclude_messages: bool = False,
    exclude_instructions: bool = False,
    exclude_tools: bool = False,
    max_length: int | None = None,
) -> Context[Deps]:
    """Create a new `Context` object with the given parameters.

    Args:
        messages : Sequence[Any] | None = None
            Initial messages to seed the context. This is any value accepted by
            `parse_context_to_pydantic_ai_messages`.
        instructions : str | Callable[[Deps | None], str] | None = None
            System-level instructions, either a string or a callable. If callable,
            it may optionally accept `deps` and should return a string.
        tools : Sequence[ToolType] | ToolType | None = None
            Tools to include in the context. This can be a sequence of functions, pydantic ai tools, builtin tools
            or toolsets.
        deps : Deps | None = None
            Reference to `deps` in `pydantic_ai.RunContext`, that can be passed to messages,
            tools and instructions.
        update : bool = True
            When `True` (default), semantic operations that consume this context will write back the full message history.
        compact_instructions : bool = False
            When `True`, all instructions are compacted into a single `SystemPromptPart`; otherwise they are split on double newlines.
        exclude_messages : bool = False
            When `True`, existing messages are not forwarded to operations.
        exclude_instructions : bool = False
            When `True`, instructions are not rendered into messages.
        exclude_tools : bool = False
            When `True`, tools are not rendered into messages.
        max_length : int | None = None
            Optional cap on the number of messages forwarded (from the tail).

    Returns:
        Context[Deps]
            A new `Context` object with the given parameters.
    """
    return Context(
        messages=messages,
        instructions=instructions,
        tools=tools,
        deps=deps,
        update=update,
        compact_instructions=compact_instructions,
        exclude_messages=exclude_messages,
        exclude_instructions=exclude_instructions,
        exclude_tools=exclude_tools,
        max_length=max_length,
    )



"""zyx.operations.validate"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, List, TypeVar, TYPE_CHECKING

from pydantic import BaseModel, Field

from .._aliases import (
    PydanticAIAgentResult,
    PydanticAIInstructions,
    PydanticAIModelSettings,
    PydanticAIUsageLimits,
)
from .._graph import (
    SemanticGraph,
    SemanticGraphDeps,
    SemanticGraphState,
    make_generate_step,
    SemanticGraphRequestTemplate,
    GraphHooks,
)
from .._types import (
    ModelParam,
    SourceParam,
    TargetParam,
    ContextType,
    ToolType,
    AttachmentType,
)
from ..result import Result
from .parse import aparse, parse

if TYPE_CHECKING:
    from .._utils._observer import Observer

__all__ = (
    "avalidate",
    "validate",
    "ConstraintViolation",
    "ConstraintViolations",
    "ValidationResult",
)

Deps = TypeVar("Deps")
Output = TypeVar("Output")


class ConstraintViolation(BaseModel):
    """A single constraint that was violated during validation."""

    constraint: str = Field(
        description="The exact constraint text that was violated."
    )
    reason: str = Field(
        description="Human-readable explanation of why the value violates this constraint."
    )


class ConstraintViolations(BaseModel):
    """Aggregated result of constraint validation.

    When every constraint is satisfied the ``violations`` list is empty.
    """

    violations: List[ConstraintViolation] = Field(
        default_factory=list,
        description=(
            "List of constraints that were violated. "
            "Empty when the value satisfies all constraints."
        ),
    )


@dataclass
class ValidationResult(Generic[Output]):
    """Result of validate when ``raise_on_error=False``.

    Contains the parsed output and constraint violations (empty when valid).
    """

    output: Output
    """The parsed output from the parse step."""

    violations: ConstraintViolations
    """Constraint validation result; ``violations.violations`` is empty when valid."""

    raw: List[PydanticAIAgentResult[Any]]
    """Underlying agent run results (parse + validate)."""


_VALIDATE_SYSTEM_PROMPT = (
    "\n[INSTRUCTION]\n"
    "- You are a constraint validator. Your ONLY task is to check whether "
    "the given value satisfies a set of constraints.\n"
    "- Evaluate every constraint carefully.\n"
    "- Return a ConstraintViolations object.\n"
    "- If ALL constraints are satisfied, return ConstraintViolations with an "
    "empty violations list.\n"
    "- If ANY constraint is violated, include an entry in the violations list "
    "with the exact constraint text and a concise reason.\n"
)


def _attach_observer_hooks(
    deps: SemanticGraphDeps[Deps, Output], operation: str
) -> None:
    observe = getattr(deps, "observe", None)
    if not observe or getattr(deps, "hooks", None) is not None:
        return
    deps.hooks = GraphHooks(
        on_run_start=lambda _ctx: observe.on_operation_start(operation),
        on_run_end=lambda _res: observe.on_operation_complete(operation),
    )


def _normalize_observer(observe: Any | None) -> Any | None:
    if not observe:
        return None
    from .._utils._observer import Observer

    if observe is True:
        return Observer()
    if isinstance(observe, Observer):
        return observe
    raise ValueError("Invalid 'observe' value. Expected bool or Observer.")


def _build_constraint_list(constraints: List[str]) -> str:
    """Format a numbered constraint list for inclusion in a system prompt."""
    lines = ["\n[CONSTRAINTS]"]
    for i, c in enumerate(constraints, 1):
        lines.append(f"  {i}. {c}")
    return "\n".join(lines)


def _build_validate_prompt(constraints: List[str], parsed_value: Any) -> str:
    """Build validation prompt that includes the parsed value and constraints."""
    from .._processing._toon import object_as_text

    constraint_text = _build_constraint_list(constraints)
    parsed_repr = object_as_text(parsed_value)
    return (
        _VALIDATE_SYSTEM_PROMPT + constraint_text + "\n"
        f"\n[PARSED VALUE]\n"
        f"The value to validate:\n{parsed_repr}\n"
        f"\nValidate this parsed value against ALL constraints listed above.\n"
    )


def _raise_on_violations(
    violations: ConstraintViolations, parsed: Any
) -> None:
    """Raise ``AssertionError`` if *violations* contains any entries.

    Includes the parsed value in the error message for context.
    """
    if violations.violations:
        from .._processing._toon import object_as_text

        formatted = "\n".join(
            f"  - {v.constraint}: {v.reason}" for v in violations.violations
        )
        parsed_repr = object_as_text(parsed)
        raise AssertionError(
            f"Constraint validation failed for parsed value:\n{parsed_repr}\n\n"
            f"Violations:\n{formatted}"
        )


async def avalidate(
    source: SourceParam,
    target: TargetParam[Output] = str,  # type: ignore[assignment]
    context: ContextType | List[ContextType] | None = None,
    *,
    constraints: List[str] | None = None,
    raise_on_error: bool = True,
    confidence: bool = False,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    attachments: AttachmentType | List[AttachmentType] | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    observe: bool | "Observer" | None = None,
) -> Result[Output] | ValidationResult[Output]:
    """Asynchronously parse a source into a target type, then validate the result against constraints.

    Args:
        source (SourceParam): The value to parse and validate.
        target (TargetParam[Output]): The type or schema to parse into. Defaults to str.
        context (ContextType | List[ContextType] | None): Optional context or conversation history. Defaults to None.
        constraints (List[str] | None): List of constraint strings to validate against. Defaults to None.
        raise_on_error (bool): If True (default), raise AssertionError if validation fails.
            If False, return a ValidationResult with any violations. Defaults to True.
        confidence (bool): If True, includes confidence scoring. Defaults to False.
        model (ModelParam): The model to use for parsing/validation. Defaults to "openai:gpt-4o-mini".
        model_settings (PydanticAIModelSettings | None): Model settings to use for the operation. Defaults to None.
        instructions (PydanticAIInstructions | None): Additional instructions for the model. Defaults to None.
        attachments (AttachmentType | List[AttachmentType] | None): Attachments to provide to the model. Defaults to None.
        tools (ToolType | List[ToolType] | None): Tools available to the model. Defaults to None.
        deps (Deps | None): Optional RunContext dependencies. Defaults to None.
        usage_limits (PydanticAIUsageLimits | None): Usage limits for the model call. Defaults to None.
        observe (bool | Observer | None): If True or provided, enables CLI observation output.

    Returns:
        Result[Output] | ValidationResult[Output]: Result[Output] if raise_on_error is True and validation passes.
            ValidationResult[Output] if raise_on_error is False (always includes violations).
    """
    from ..targets import Target

    observe_handler = _normalize_observer(observe)
    if observe_handler:
        observe_handler.on_operation_start("validate")

    try:
        _constraints: List[str] = constraints or []
        if isinstance(target, Target) and getattr(target, "constraints", None):
            _constraints = target.constraints or _constraints  # type: ignore[assignment]

        parse_result = await aparse(
            source=source,
            target=target,
            context=context,
            confidence=confidence,
            model=model,
            model_settings=model_settings,
            instructions=instructions,
            tools=tools,
            deps=deps,
            attachments=attachments,
            usage_limits=usage_limits,
            observe=observe_handler,
            stream=False,
        )
        parsed_value = parse_result.output

        if _constraints:
            validate_deps = SemanticGraphDeps.prepare(
                target=ConstraintViolations,
                source=None,
                model=model,
                model_settings=model_settings,
                context=None,
                instructions=_build_validate_prompt(
                    _constraints, parsed_value
                ),
                tools=None,
                deps=deps,
                attachments=attachments,
                usage_limits=usage_limits,
                observe=observe_handler,
            )
            _attach_observer_hooks(validate_deps, "validate")
            validate_state = SemanticGraphState.prepare(deps=validate_deps)
            validate_request = SemanticGraphRequestTemplate(
                include_source_context=False,
            )
            validate_graph = SemanticGraph(
                steps=[
                    make_generate_step(
                        validate_request,
                        update_output=False,
                    )
                ],
                state=validate_state,
                deps=validate_deps,
            )
            validate_result = await validate_graph.run()
            violations = validate_result.output
            if (
                isinstance(violations, ConstraintViolations)
                and violations.violations
            ):
                if raise_on_error:
                    _raise_on_violations(violations, parsed_value)
                return ValidationResult(
                    output=parsed_value,
                    violations=violations,
                    raw=parse_result.raw + validate_result.raw,
                )
            if raise_on_error:
                return parse_result
            return ValidationResult(
                output=parsed_value,
                violations=violations,
                raw=parse_result.raw + validate_result.raw,
            )

        else:
            return parse_result
    finally:
        if observe_handler:
            observe_handler.on_operation_complete("validate")


def validate(
    source: SourceParam,
    target: TargetParam[Output] = str,  # type: ignore[assignment]
    context: ContextType | List[ContextType] | None = None,
    *,
    constraints: List[str] | None = None,
    raise_on_error: bool = True,
    confidence: bool = False,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    attachments: AttachmentType | List[AttachmentType] | None = None,
    instructions: PydanticAIInstructions | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    observe: bool | "Observer" | None = None,
) -> Result[Output] | ValidationResult[Output]:
    """Synchronously parse a source into a target type, then validate the result against constraints.

    Args:
        source (SourceParam): The value to parse and validate.
        target (TargetParam[Output]): The type or schema to parse into. Defaults to str.
        context (ContextType | List[ContextType] | None): Optional context or conversation history. Defaults to None.
        constraints (List[str] | None): List of constraint strings to validate against. Defaults to None.
        raise_on_error (bool): If True (default), raise AssertionError if validation fails.
            If False, return a ValidationResult with any violations. Defaults to True.
        confidence (bool): If True, includes confidence scoring. Defaults to False.
        model (ModelParam): The model to use for parsing/validation. Defaults to "openai:gpt-4o-mini".
        model_settings (PydanticAIModelSettings | None): Model settings to use for the operation. Defaults to None.
        instructions (PydanticAIInstructions | None): Additional instructions for the model. Defaults to None.
        attachments (AttachmentType | List[AttachmentType] | None): Attachments to provide to the model. Defaults to None.
        tools (ToolType | List[ToolType] | None): Tools available to the model. Defaults to None.
        deps (Deps | None): Optional RunContext dependencies. Defaults to None.
        usage_limits (PydanticAIUsageLimits | None): Usage limits for the model call. Defaults to None.
        observe (bool | Observer | None): If True or provided, enables CLI observation output.

    Returns:
        Result[Output] | ValidationResult[Output]: Result[Output] if raise_on_error is True and validation passes.
            ValidationResult[Output] if raise_on_error is False (always includes violations).
    """
    from ..targets import Target

    observe_handler = _normalize_observer(observe)
    if observe_handler:
        observe_handler.on_operation_start("validate")

    try:
        _constraints: List[str] = constraints or []
        if isinstance(target, Target) and getattr(target, "constraints", None):
            _constraints = target.constraints or _constraints  # type: ignore[assignment]

        parse_result = parse(
            source=source,
            target=target,
            context=context,
            confidence=confidence,
            model=model,
            model_settings=model_settings,
            attachments=attachments,
            instructions=instructions,
            tools=tools,
            deps=deps,
            usage_limits=usage_limits,
            observe=observe_handler,
            stream=False,
        )
        parsed_value = parse_result.output

        if _constraints:
            validate_deps = SemanticGraphDeps.prepare(
                target=ConstraintViolations,
                source=None,
                model=model,
                model_settings=model_settings,
                context=None,
                instructions=_build_validate_prompt(
                    _constraints, parsed_value
                ),
                tools=None,
                deps=deps,
                attachments=attachments,
                usage_limits=usage_limits,
                observe=observe_handler,
            )
            _attach_observer_hooks(validate_deps, "validate")
            validate_state = SemanticGraphState.prepare(deps=validate_deps)
            validate_request = SemanticGraphRequestTemplate(
                include_source_context=False,
            )
            validate_graph = SemanticGraph(
                steps=[
                    make_generate_step(
                        validate_request,
                        update_output=False,
                    )
                ],
                state=validate_state,
                deps=validate_deps,
            )
            validate_result = validate_graph.run_sync()
            violations = validate_result.output
            if (
                isinstance(violations, ConstraintViolations)
                and violations.violations
            ):
                if raise_on_error:
                    _raise_on_violations(violations, parsed_value)
                return ValidationResult(
                    output=parsed_value,
                    violations=violations,
                    raw=parse_result.raw + validate_result.raw,
                )
            if raise_on_error:
                return parse_result
            return ValidationResult(
                output=parsed_value,
                violations=violations,
                raw=parse_result.raw + validate_result.raw,
            )

        else:
            return parse_result
    finally:
        if observe_handler:
            observe_handler.on_operation_complete("validate")
